//----------------------------------------------------------------------------
// Copyright 2018-2024 Facilis Technology Inc.
//----------------------------------------------------------------------------
#include "ptsys.h"
#include "ptalarm.h"
#include "ptremote.h"
#include "ptsst.h"
#include "ptjson.h"
#include "facutil.h"
#include "EventDebouncer.h"
#include "facconfig.h"
#include "HubAPIServer.h"
#define TRACE_CHANNEL g_FacConfigTc
#include "pttraces.h"
#include "fshub.rc"

szc_t TimestampToSz( malloc_i* resMa, int64_t timestamp)
{
    mp_c mp;
    int year, month, day, hour, minute, second;
    file_time_unpack( timestamp, &year, &month, &day, &hour, &minute, &second, nullptr/*msec*/);
    return ssformat( resMa, "%04d-%02d-%02dT%02d:%02d:%02d",
        year  ,
        month ,
        day   ,
        hour  ,
        minute,
        second);
}

int64_t SzToTimestamp( szc_t timestampSz)
{
    int64_t timestamp = 0;
    if (!timestampSz.isempty())
    {
        int year, month, day, hour, minute, second;
        ss8totime( &year, &month, &day, &hour,
            &minute, &second, nullptr/*msec*/, timestampSz);
        timestamp = file_time_pack( year, month, day, hour, minute, second, 0/*msec*/);
    }
    return timestamp;
}

struct FacConfig_EventDebouncerOwner
:
    IEventDebouncerOwner
{
    virtual void BusyRetain() = 0;
    virtual void BusyRelease() = 0;

    int m_RefCount;

    void IEventDebouncerOwner_Retain()
    {
        ASSERT( m_RefCount >= 0);
        if (sync_inc( &m_RefCount) == 1)
        {
            BusyRetain();
        }
    }

    void IEventDebouncerOwner_Release()
    {
        ASSERT( m_RefCount > 0);
        if (sync_dec( &m_RefCount) == 0)
        {
            BusyRelease();
        }
    }
};

struct FacConfig
:
    IFacConfig,
    FacConfig_EventDebouncerOwner
{
    IFacConfigOwner* m_Owner;
    IHubAPI* m_Hub;
    pin_c m_CancelPin;
    pin_c m_CommitPin;
    pin_c m_ReplicationPin;
    int m_BusyCount;
    int m_RefCount;
    sz_c m_GroupName;
    sz_c m_LocalId;
    sz_c m_StorageLocation;
    IFacOutput* m_Output;
    bool m_FailedLoad;
    blob_i* m_Secret;
    sz_c m_ServerName;
    bool m_Started;
    IEventDebouncer* m_Debouncer;

    int m_ReplicationInProgress;
    int m_ReplicationNeeded;

    sync_c m_MainLock;

    struct Peer;
    struct RepChange;
    rblink_t* m_CenterPeer;
    Peer* m_MasterPeer;
    sz_c m_Priority;

    RepChange* m_RepChangeListFirst;
    RepChange* m_RepChangeListLast;

    struct FacKey;
    struct FacVal;
    struct FacItem;
    FacKey* root;

    void outf( szc_t format, ...)
    {
        va_list args;
        va_start(args, format);
        m_Output->outfv( args, format);
        va_end(args);
    }

    void tracef( szc_t inFormat, ...)
    {
        mp_c mp;
        szc_t format = ssformat( &mp, "%s: %s\n", m_ServerName.c_str, inFormat);
        va_list args;
        va_start(args, inFormat);
        m_Output->tracefv( args, format);
        va_end(args);
    }

        // This constructor expects zero-initialized memory.
    void* operator new( size_t size, void* mem) { return mem; }
    FacConfig( IFacOutput* output, IFacConfigOwner* owner, IHubAPI* hub, szc_t storageLocation)
    :
        m_Output( output),
        m_Owner( owner),
        m_Hub( hub),
        m_StorageLocation( storageLocation),
        m_RefCount( 1),
        m_BusyCount( 1),
        m_Debouncer( EventDebouncerFactory( this))
    {
    }

    ~FacConfig()
    {
        RepChange* repChange;
        while ((repChange = PopRepChange()) != nullptr)
        {
            repChange->Release();
        }
        Empty();
        ASSERT( !m_RefCount);
        ASSERT( !m_BusyCount);
        if (m_Secret) m_Secret->release();
        if (root) root->Remove();
        m_Debouncer->Release();
    }

    void /*FacConfig_EventDebouncerOwner*/BusyRetain() override
    {
        ASSERT( m_BusyCount > 0);
        sync_inc( &m_BusyCount);
    }

    void /*FacConfig_EventDebouncerOwner*/BusyRelease() override
    {
        ASSERT( m_BusyCount > 0);
        if (sync_dec( &m_BusyCount) == 0)
        {
            ASSERT( m_RefCount == 0);
            this->~FacConfig();
            free( this);
        }
    }

    void /*IFacConfig*/Retain() override
    {
        ASSERT( m_RefCount > 0);
        if (sync_inc( &m_RefCount) == 1)
        {
            BusyRetain();
        }
    }

    void /*IFacConfig*/Cancel() override
    {
        m_CancelPin.signal();
        m_Debouncer->Cancel();
    }

    void /*IFacConfig*/Stop() override
    {
    }

    void /*IFacConfig*/Release() override
    {
        ASSERT(m_RefCount > 0);
        if (sync_dec( &m_RefCount) == 0)
        {
            Cancel();
            BusyRelease();
        }
    }

    pin_i* CancelPin()
    {
        return &m_CancelPin;
    }

    pin_i* CommitPin()
    {
        return &m_CommitPin;
    }

    pin_i* ReplicationPin()
    {
        return &m_ReplicationPin;
    }

    void CommitNeeded()
    {
        m_CommitPin.pulse();
    }

    void ReplicationNeeded()
    {
        m_ReplicationPin.pulse();
    }

    void GetLocalIdPriority( malloc_i* resMa, sz_t* outLocalId, sz_t* outPriority)
    {
        m_MainLock.lock();
        if (outLocalId) *outLocalId = sscpy( resMa, m_LocalId);
        if (outPriority) *outPriority = sscpy( resMa, m_Priority);
        m_MainLock.unlock();
    }

    bool IsLocalPeer( szc_t peerId)
    {
        m_MainLock.lock();
        bool res = sscmp( m_LocalId, peerId) == 0;
        m_MainLock.unlock();
        return res;
    }

    struct ValLink
    :
        rblink_t
    {
        FacVal* val;

        static int ValLinkCompare(void* name_, rblink_t* valLink_)
        {
            szc_t name( SCAST( const char*, name_));
            ValLink* itemLink = SCAST( ValLink*, valLink_);
            return ss8asciicmpi( name, itemLink->val->m_Name);
        }
    };

    void /*FacConfig_EventDebouncerOwner*/IEventDebouncerOwner_Signal( szc_t path) override
    {
        mp_c mp;
        sz_t scratch = sscpy( &mp, path);
        sz_t itemName = nullptr;
        char* sep = ssrchr( scratch, '/');
        if (sep)
        {
            *sep = 0;
            itemName = sep + 1;
        }
        m_Owner->IFacConfigOwner_ProcessNotify( scratch/*parentKeyName*/, itemName);
    }

    struct Filter
    {
        int m_RefCount;
        rblink_t m_FilterLink;
        rblink_t* m_CenterValLink;
        sz_c m_FilterKey;
        sz_c m_FilterData;
        sync_c m_Lock;
        bool m_Initialized;
        FacKey* m_Parent;
        uint64_t m_ChangeId;
        int m_FilterType;

        static const int valFilterType = 1;
        static const int jsonFilterType = 2;

        void* operator new(size_t size, void* mem) { return mem; }
        Filter( FacKey* parent, int filterType, szc_t filterKey, szc_t filterData)
        :
            m_Parent( parent),
            m_FilterType( filterType),
            m_FilterKey( filterKey),
            m_FilterData( filterData),
            m_RefCount( 1)
        {
        }

        ~Filter()
        {
            ASSERT( !m_RefCount);
            ASSERT( !m_CenterValLink); // Should be cleaned up by now
            Empty();
        }

        static int FilterCompare(void* params_, rblink_t* filter_)
        {
            szc_t* params = SCAST( szc_t*, params_);
            int filterType = SCAST( int, sstoint( params[0], SIZE_MAX, nullptr));
            szc_t filterKey = params[1];
            szc_t filterData = params[2];
            Filter* filter = containerof( Filter, m_FilterLink, filter_);
            int diff = filterType - filter->m_FilterType;
            if (diff) return diff;
            diff = ss8asciicmpi( filterKey, filter->m_FilterKey);
            if (diff) return diff;
            return ss8asciicmpi( filterData, filter->m_FilterData);
        }

        void Retain()
        {
            ASSERT( m_RefCount > 0);
            sync_inc( &m_RefCount);
        }

        void Release()
        {
            ASSERT( m_RefCount > 0);
            if (sync_dec( &m_RefCount) == 0)
            {
                this->~Filter();
                free( this);
            }
        }

        void AddVal( FacVal* val)
        {
            rblink_t* link;
            rblink_t* prev;
            val->FacItem_Retain();
            ValLink* valLink = t_calloc( ValLink);
            m_Lock.lock();
            link = rblink_find_2( m_CenterValLink, ValLink::ValLinkCompare, val->m_Name.c_str, &prev);
            if (!link)
            {
                valLink->val = val;
                m_ChangeId++;
                rblink_insert( &m_CenterValLink, prev, valLink);
                valLink = nullptr;
            }
            m_Lock.unlock();
            if (valLink) free( valLink);
        }

        void RemoveVal( szc_t name)
        {
            rblink_t* link;
            ValLink* valLink = nullptr;
            m_Lock.lock();
            link = rblink_find( m_CenterValLink, ValLink::ValLinkCompare, name.c_str);
            if (link)
            {
                rblink_remove( &m_CenterValLink, link);
                valLink = SCAST( ValLink*, link);
                m_ChangeId++;
            }
            m_Lock.unlock();
            if (valLink)
            {
                valLink->val->FacItem_Release();
                free( valLink);
            }
        }

        void Empty()
        {
            mp_c mp;
            ValLink* valLink;
            rblink_t* link;
            m_Lock.lock();
            while ((link = m_CenterValLink) != nullptr)
            {
                m_ChangeId++;
                rblink_remove( &m_CenterValLink, link);
                m_Lock.unlock();
                valLink = SCAST( ValLink*, link);
                valLink->val->RemoveFilterLinkUnderLock( m_FilterType, m_FilterKey, m_FilterData);
                free( valLink);
                m_Lock.lock();
            }
            m_Lock.unlock();
        }

        size_t/*count*/ GetValList( malloc_i* resMa, FacKeyVal*** outList)
        {
            size_t valCount = 0;
            size_t count = 0;
            FacKeyVal** list;
            m_Lock.lock();
            rblink_t* link = nullptr;
            while ((link = rblink_next( m_CenterValLink, link)) != nullptr)
            {
                valCount++;
            }
            list = ma_n_alloc( resMa, FacKeyVal*, valCount + 1);
            link = nullptr;
            while ((link = rblink_next( m_CenterValLink, link)) != nullptr)
            {
                ValLink* valLink = SCAST( ValLink*, link);
                FacKeyVal* keyVal = ma_t_alloc( resMa, FacKeyVal);
                keyVal->name = sscpy( resMa, valLink->val->m_Name);
                keyVal->val = sscpy( resMa, valLink->val->m_Val);
                list[count] = keyVal;
                count++;
            }
            m_Lock.unlock();
            if (outList) *outList = list;
            ASSERT( count == valCount);
            return count;
        }

        IFacList* GetValList()
        {
                // TODO: Implement it generation
            return new(t_calloc(FacList)) FacFilterList( this);
        }
    };

    static Filter* FilterFactory( FacKey* parent, int filterType, szc_t filterKey, szc_t filterData)
    {
        mp_c mp;
        szc_t params[3] = { ssformat( &mp, "%i", filterType), filterKey, filterData};
        Filter* filter = new(t_calloc(FacKey)) Filter( parent, filterType, filterKey, filterData);
        Filter* extra = nullptr;
        rblink_t* link;
        rblink_t* prev;
        parent->m_KeyLock.lock();
        link = rblink_find_2( parent->m_CenterFilter, Filter::FilterCompare, params, &prev);
        if (link)
        {
            extra = filter;
            filter = containerof( Filter, m_FilterLink, link);
        }
        else
        {
            rblink_insert( &parent->m_CenterFilter, prev, &filter->m_FilterLink);
            while ((link = rblink_next( parent->m_Center, link)) != nullptr)
            {
                FacItem* item = containerof( FacItem, m_Link, link);
                if (item->m_Type == FacItem::valType)
                {
                    const char* valSz = nullptr;
                    FacVal* val = SCAST( FacVal*, item);
                    if (!val->m_Json)
                    {
                        val->m_Json = ptjson_import( &val->m_JsonMp, nullptr, val->m_Val, sslen( val->m_Val));
                    }
                    valSz = ptjson_get_sz(&mp, val->m_Json, filter->m_FilterKey);
                    if (valSz && ss8asciicmpi( valSz, filter->m_FilterData) == 0)
                    {
                        val->AddFilterLinkUnderLock( filter);
                    }
                }
            }
        }
        filter->Retain();
        parent->m_KeyLock.unlock();
        if (extra) extra->Release();
        return filter;
    }

    struct FacListBase
    {
        uint64_t m_LastChangeId;
        sz_c m_LastKeyName;
        rblink_t* m_LastLink = nullptr;
    };

    struct FacList
    :
        FacListBase,
        IFacList
    {
        FacKey* m_Key;

        void* operator new(size_t size, void* mem) { return mem; }
        FacList( FacKey* key)
        :
            m_Key( key)
        {
            m_Key->Retain();
        }

        void /*IFacList*/Release()
        {
            m_Key->Release();
            this->~FacList();
            free( this);
        }

        void /*IFacList*/Reset()
        {
            m_LastChangeId = 0;
            m_LastKeyName = nullptr;
            m_LastLink = nullptr;
        }

        FacItem* Next( int itemType)
        {
            FacItem* item = nullptr;
            m_Key->m_KeyLock.lock();
            if (m_LastChangeId != m_Key->m_ChangeId)
            {
                rblink_find_2( m_Key->m_Center, FacItem::Compare, m_LastKeyName.c_str, &m_LastLink);
            }
            while( (m_LastLink = rblink_next( m_Key->m_Center, m_LastLink)) != nullptr)
            {
                item = containerof( FacItem, m_Link, m_LastLink);
                ASSERT( item->m_Type >= FacItem::folderType && item->m_Type <= FacItem::whiteoutType);
                if (item->m_Type == itemType)
                {
                    m_LastKeyName = item->m_Name;
                    break;
                }
                else
                {
                    item = nullptr;
                }
            }
            m_LastChangeId = m_Key->m_ChangeId;
            if (item) item->FacItem_Retain();
            m_Key->m_KeyLock.unlock();
            return item;
        }

        IFacKey* /*IFacList*/NextKey()
        {
            FacItem* item = Next( FacItem::folderType);
            FacKey* key = nullptr;
            if (item)
            {
                key = SCAST( FacKey*, item);
            }
            return key;
        }

        FacKeyVal* /*IFacList*/NextVal( malloc_i* resMa)
        {
            FacKeyVal* valKey = nullptr;
            FacItem* item = Next( FacItem::valType);
            if (item)
            {
                FacVal* val = SCAST( FacVal*, item);

                valKey = ma_t_alloc( resMa, FacKeyVal);
                valKey->name = sscpy( resMa, val->m_Name);
                valKey->val = sscpy( resMa, val->m_Val);

                item->FacItem_Release();
            }
            return valKey;
        }
    };

    struct FacFilterList
    :
        FacListBase,
        IFacList
    {
        Filter* m_Filter;

        void* operator new(size_t size, void* mem) { return mem; }
        FacFilterList( Filter* filter)
        :
            m_Filter( filter)
        {
            m_Filter->Retain();
        }

        void /*IFacList*/Release()
        {
            m_Filter->Release();
            this->~FacFilterList();
            free( this);
        }

        void /*IFacList*/Reset()
        {
            m_LastChangeId = 0;
            m_LastKeyName = nullptr;
            m_LastLink = nullptr;
        }

        FacItem* Next()
        {
            FacVal* val = nullptr;
            m_Filter->m_Lock.lock();
            if (m_LastChangeId != m_Filter->m_ChangeId)
            {
                rblink_t* link = rblink_find_2( m_Filter->m_CenterValLink, ValLink::ValLinkCompare, m_LastKeyName.c_str, &m_LastLink);
                if (link) m_LastLink = link;
            }
            m_LastLink = rblink_next( m_Filter->m_CenterValLink, m_LastLink);
            if (m_LastLink != nullptr)
            {
                ValLink* valLink = SCAST( ValLink*, m_LastLink);
                val = valLink->val;
                ASSERT( val->m_Type == FacItem::valType);
                m_LastKeyName = val->m_Name;
            }
            m_LastChangeId = m_Filter->m_ChangeId;
            if (val) val->FacItem_Retain();
            m_Filter->m_Lock.unlock();
            return val;
        }

        IFacKey* /*IFacList*/NextKey()
        {
                // Unused
            return nullptr;
        }

        FacKeyVal* /*IFacList*/NextVal( malloc_i* resMa)
        {
            FacKeyVal* valKey = nullptr;
            FacItem* item = Next();
            if (item)
            {
                FacVal* val = SCAST( FacVal*, item);

                valKey = ma_t_alloc( resMa, FacKeyVal);
                valKey->name = sscpy( resMa, val->m_Name);
                valKey->val = sscpy( resMa, val->m_Val);

                item->FacItem_Release();
            }
            return valKey;
        }
    };

    struct ItemNode
    {
        int64_t timestamp;
        sz_t name;
        sz_t val;
        int type;
    };

    static ItemNode* ItemNodeFactory( malloc_i* resMa, FacItem* item)
    {
        ItemNode* node = ma_t_alloc( resMa, ItemNode);
        FacVal* val;
        switch (item->m_Type)
        {
            case FacItem::folderType:
            case FacItem::whiteoutType:
                break;
            case FacItem::valType:
                val = SCAST( FacVal*, item);
                node->val = sscpy( resMa, val->m_Val);
                break;
            default:
                ASSERT( "Unsupported item type" && false);
                break;
        }
        node->name = sscpy( resMa, item->m_Name);
        node->timestamp = item->m_Timestamp;
        node->type = item->m_Type;
        ASSERT( node->type >= FacItem::folderType && node->type <= FacItem::whiteoutType);
        return node;
    }

    struct FacItem
    {
        FacKey* m_Parent;
        int m_Type;
        rblink_t m_Link;
        sz_c m_Name;
        uint64_t m_ChangeId;
        int/*bool*/ m_Removed;

        int64_t m_Timestamp;
        int m_RefCount;

        static const int folderType = 1;
        static const int valType = 2;
        static const int whiteoutType = 3;

        virtual ~FacItem()
        {
            ASSERT( !m_RefCount);
            if (m_Parent) m_Parent->Release();
            ASSERT( m_Removed);
        }

        virtual void Remove() = 0;

        void FacItem_Retain()
        {
            ASSERT( m_RefCount > 0);
            sync_inc( &m_RefCount);
        }

        void FacItem_Release()
        {
            ASSERT( m_RefCount > 0);
            if (sync_dec( &m_RefCount) == 0)
            {
                this->~FacItem();
                free( this);
            }
        }

        void UpdateFilterChangeIdUnderLock()
        {
            m_ChangeId++;
        }

        static int Compare(void* name_, rblink_t* index_)
        {
            szc_t name( SCAST( const char*, name_));
            FacItem* item = containerof( FacItem, m_Link, index_);
            return ss8asciicmpi( name, item->m_Name);
        }
    };

    struct FacVal
    :
        FacItem
    {
        sz_c m_Val;
        int m_FilterType;
        mp_c m_JsonMp;
        json_t* m_Json;
        rblink_t* m_CenterActiveFilter; // From parent object

        struct FilterLink
        :
            rblink_t
        {
            Filter* filter;

            static int FilterLinkCompare(void* params_, rblink_t* filterLink_)
            {
                szc_t* params = SCAST( szc_t*, params_);
                szc_t filterKey = params[0];
                szc_t filterData = params[1];
                FilterLink* filterLink = SCAST( FilterLink*, filterLink_);
                int diff = ss8asciicmpi( filterKey, filterLink->filter->m_FilterKey);
                if (diff) return diff;
                return ss8asciicmpi( filterData, filterLink->filter->m_FilterData);
            }
        };

        FacVal( szc_t name, szc_t val, int64_t timestamp)
        :
            m_Val( val)
        {
            m_Timestamp = timestamp;
            json_t* json = ptjson_import( &m_JsonMp, nullptr, m_Val, sslen( m_Val));
            m_FilterType = Filter::valFilterType;
            if (json)
            {
                m_FilterType = Filter::jsonFilterType;
                m_Json = json;
            }
            m_Name = name;
            m_RefCount = 1;
            m_Type = FacItem::valType;
        }

        ~FacVal()
        {
            ASSERT( !m_CenterActiveFilter);
        }

        void Remove()
        {
            if (m_Parent)
            {
                ASSERT( !rblink_is_linked( m_Parent->m_Center, &m_Link));
            }
            if (!m_Removed)
            {
                m_Removed = true;
                FilterLink* filterLink;
                rblink_t* link;
                while ((link = m_CenterActiveFilter) != nullptr)
                {
                    rblink_remove( &m_CenterActiveFilter, link);
                    filterLink = SCAST( FilterLink*, link);
                    filterLink->filter->RemoveVal( m_Name);
                    filterLink->filter->Release();
                    free( filterLink);
                }
                //ASSERT( m_RefCount == 1);
                FacItem_Release();
            }
            else
            {
                ASSERT( "Removed twice" && false);
            }
        }

            // Always called under lock of parent
        int/*changed*/ SetValUnderParentLock( szc_t val, int64_t timestamp)
        {
            mp_c mp;
            json_t* json = ptjson_import( &mp, nullptr, val, sslen(val));
            int changed = 0;
            struct Box
            {
                FilterLink* link;
                Box* next;
            };
            Box* first = nullptr;
            Box* box;
            ASSERT( sync_is_locked( &m_Parent->m_KeyLock));
            if (ss8asciicmpi( val, m_Val) != 0)
            {
                if (m_Json)
                {
                        // Find active filters to remove later
                    rblink_t* link = nullptr;
                    while ((link = rblink_next( m_CenterActiveFilter, link)) != nullptr)
                    {
                        FilterLink* filterLink = SCAST( FilterLink*, link);
                        Filter* filter = filterLink->filter;
                        if (ss8asciicmpi( ptjson_get_sz( &mp, m_Json, filter->m_FilterKey), filter->m_FilterData) != 0)
                        {
                            box = ma_t_alloc( &mp, Box);
                            box->link = filterLink;
                            box->next = first;
                            first = box;
                        }
                    }
                }
                changed = 1;
                m_Val = val;
                m_JsonMp.reset();
                m_Json = ptjson_new_copy( &m_JsonMp, json);
                if (m_Json)
                {
                    m_FilterType = Filter::jsonFilterType;
                }
                else
                {
                    m_FilterType = Filter::valFilterType;
                }
            }
            m_Timestamp = timestamp;
            UpdateFilterChangeIdUnderLock();
            for (box = first; box; box = box->next)
            {
                    // Not ideal to release this under lock
                box->link->filter->RemoveVal( m_Name);
                box->link->filter->Release();
                rblink_remove( &m_CenterActiveFilter, box->link);
                free( box->link);
            }
            return changed;
        }

            // Always called under lock of parent
        sz_t GetVal( malloc_i* resMa)
        {
            ASSERT( sync_is_locked( &m_Parent->m_KeyLock));
            return sscpy( resMa, m_Val);
        }

        void AddFilterLinkUnderLock( Filter* filter)
        {
            mp_c mp;
            ASSERT( m_Parent->m_KeyLock.is_locked());
            szc_t params[2] = { filter->m_FilterKey, filter->m_FilterData};
            rblink_t* link;
            rblink_t* prev;
            ASSERT( !m_Removed);
            link = rblink_find_2( m_CenterActiveFilter, FilterLink::FilterLinkCompare, params, &prev);
            if (!link)
            {
                FilterLink* filterLink = t_calloc( FilterLink);
                filterLink->filter = filter;
                filterLink->filter->Retain();
                rblink_insert( &m_CenterActiveFilter, prev, filterLink);
                filter->AddVal( this);
            }
        }

        void RemoveFilterLinkUnderLock( int filterType, szc_t filterKey, szc_t filterData)
        {
            mp_c mp;
            ASSERT( m_Parent->m_KeyLock.is_locked());
            szc_t params[2] = { filterKey, filterData};
            FilterLink* filterLink = nullptr;
            rblink_t* link = rblink_find( m_CenterActiveFilter, FilterLink::FilterLinkCompare, params);
            if (link)
            {
                rblink_remove( &m_CenterActiveFilter, link);
                filterLink = SCAST( FilterLink*, link);
                filterLink->filter->RemoveVal( m_Name);
                filterLink->filter->Release();
                free( filterLink);
            }
        }
    };

    static FacVal* ValFactory( szc_t name, szc_t val, int64_t timestamp)
    {
        return new(t_calloc(FacVal)) FacVal( name, val, timestamp);
    }

    struct FacWhiteout
    :
        FacItem
    {
        FacWhiteout( szc_t name, int64_t timestamp)
        {
            m_Timestamp = timestamp;
            m_Name = name;
            m_RefCount = 1;
            m_Type = FacItem::whiteoutType;
        }

        ~FacWhiteout()
        {
        }

        void Remove()
        {
            if (m_Parent)
            {
                ASSERT( !rblink_is_linked( m_Parent->m_Center, &m_Link));
            }
            if (!m_Removed)
            {
                m_Removed = true;
                ASSERT( m_RefCount == 1);
                FacItem_Release();
            }
            else
            {
                ASSERT( "Removed twice" && false);
            }
        }

        static bool Expired( int64_t timestamp)
        {
                // 30 days in 100nsec
            static int64_t whiteoutExpiration = 25920000000000L;
            int64_t now = time_100nsec();

            int64_t diff = now > timestamp ? now - timestamp : timestamp - now;
            return diff > whiteoutExpiration;
        }
    };

    static FacWhiteout* WhiteoutFactory( szc_t name, int64_t timestamp)
    {
        return new(t_calloc(FacWhiteout)) FacWhiteout( name, timestamp);
    }

    static size_t InitKeyPathList( malloc_i* resMa, sz_t** outKeyPathList, sz_t* parentKeyPathList, size_t parentKeyPathCount, szc_t name)
    {
        size_t keyPathCount = parentKeyPathCount + 1;
        sz_t* keyPathList = ma_n_alloc( resMa, sz_t, keyPathCount + 1);
        size_t i;
        for (i = 0; i < parentKeyPathCount; i++)
        {
            keyPathList[i] = sscpy( resMa, parentKeyPathList[i]);
        }
        ASSERT( i == keyPathCount - 1);
        keyPathList[i] = sscpy( resMa, name);
        *outKeyPathList = keyPathList;
        return keyPathCount;
    }

    struct FacKey
    :
        FacItem,
        IFacKey
    {
        FacConfig* m_Config;

        mp_c m_KeyPathMp;
        sz_t* m_KeyPathList;
        size_t m_KeyPathCount;
        rblink_t* m_Center;
        rblink_t* m_CenterFilter;
        sync_c m_KeyLock;
        mux_c m_JsonKeyMutex;
        bool m_NotifyOwnerOnChange;

        void* operator new(size_t size, void* mem) { return mem; }
        FacKey( FacConfig* config, szc_t name)
        :
            m_Config( config)
        {
            m_Name = name;
            m_RefCount = 1;
            m_Type = FacItem::folderType;
        }

        ~FacKey()
        {
            ASSERT( !m_CenterFilter);
        }

        void /*IFacKey*/Retain() override
        {
            FacItem_Retain();
        }

        void /*IFacKey*/Release() override
        {
            FacItem_Release();
        }

        void Remove() override
        {
            bool removed = false;
            m_KeyLock.lock();
            if (!m_Removed)
            {
                m_Removed = removed = true;
            }
            else
            {
                ASSERT( "Removed twice" && false);
            }
            m_KeyLock.unlock();
            if (removed)
            {
                Empty();
                Release();
            }
        }

        void InitKeyPath()
        {
            sz_t* parentKeyPathList = m_Parent ? m_Parent->m_KeyPathList : nullptr;
            size_t parentKeyPathCount = m_Parent ? m_Parent->m_KeyPathCount : 0;
            sz_t* keyPathList;
            m_KeyPathCount = InitKeyPathList( &m_KeyPathMp, &keyPathList, parentKeyPathList, parentKeyPathCount, m_Name);
            m_KeyPathList = keyPathList;
        }

       // KeyNode* GetKeyNode( malloc_i* resMa)
       // {
       //     KeyNode* node = ma_t_alloc( resMa, KeyNode);
       //     node->timestamp = m_Timestamp;
       //     node->keyPath = sscpy( resMa, m_KeyPath);
       //     return node;
       // }

        FacItem* RetainItem( szc_t name, int itemType)
        {
            rblink_t* link;
            FacItem* item = nullptr;
            m_KeyLock.lock();
            link = rblink_find( m_Center, FacItem::Compare, name);
            if (link)
            {
                item = containerof( FacItem, m_Link, link);
                if (item->m_Type == itemType)
                {
                    item->FacItem_Retain();
                }
                else
                {
                    item = nullptr;
                }
            }
            m_KeyLock.unlock();
            return item;
        }

        FacVal* RetainVal( szc_t name)
        {
            FacVal* val = nullptr;
            FacItem* item = RetainItem( name, FacItem::valType);
            if (item)
            {
                val = SCAST( FacVal*, item);
            }
            return val;
        }

        FacKey* RetainKey( szc_t name)
        {
            FacKey* key = nullptr;
            FacItem* item = RetainItem( name, FacItem::folderType);
            if (item)
            {
                key = SCAST( FacKey*, item);
            }
            return key;
        }

        void UpdateTimestamp( int64_t timestamp, bool propagateTimestamp)
        {
            FacKey* key = this;
            while (key)
            {
                key->m_KeyLock.lock();
                key->m_Timestamp = timestamp;
                key->m_KeyLock.unlock();
                key = key->m_Parent;
                if (!propagateTimestamp)
                {
                    break;
                }
TRAP;
            }
        }

        void UpdateTimestamp( int64_t timestamp)
        {
            UpdateTimestamp( timestamp, true/*propagateTimestamp*/);
        }

        IFacKey* /*IFacKey*/OpenKey( szc_t name) override
        {
            return RetainKey( name);
        }

        FacKey* CreateKey( szc_t name, int64_t timestamp, bool propagateTimestamp, bool overwriteTimestamp, szc_t owningPeerId)
        {
            mp_c mp;
            FacItem* extra = nullptr;
            FacItem* removeItem = nullptr;
            FacKey* key = m_Config->KeyFactory( name);
            bool shouldCommit = false;
            rblink_t* collision;
            rblink_t* prev;
            m_KeyLock.lock();
            if (!m_Removed)
            {
                collision = rblink_find_2( m_Center, FacItem::Compare, name, &prev);
                if (collision)
                {
                    FacItem* item = containerof( FacItem, m_Link, collision);
                    if (item->m_Type == FacItem::folderType)
                    {
                        extra = key;
                        key = SCAST( FacKey*, item);
                    }
                    else
                    {
                        removeItem = item;
                        rblink_remove( &m_Center, collision);
                        collision = nullptr;
                    }
                }
                if (!collision)
                {
                    shouldCommit = true;
                    rblink_insert( &m_Center, prev, &key->m_Link);
                    UpdateFilterChangeIdUnderLock();
                        // Child reference
                    Retain();
                    key->m_Parent = this;
                    key->InitKeyPath();
                }
                key->Retain();
                ASSERT( key->m_RefCount >= 2);
            }
            else
            {
                extra = key;
                key = nullptr;
            }
            m_KeyLock.unlock();
            if (removeItem)
            {
                if (removeItem->m_Type == FacItem::folderType)
                {
                    Release();
                }
                removeItem->Remove();
            }
            if (extra)
            {
                ASSERT( extra->m_RefCount == 1);
                extra->Remove();
            }
                // This means it changed
            if ((shouldCommit || overwriteTimestamp) && key)
            {
                key->UpdateTimestamp( timestamp, propagateTimestamp);
                m_Config->RepChangeFactory( owningPeerId, FacItem::folderType, key, name, nullptr/*val*/, timestamp);
                shouldCommit = true;
            }
            if (shouldCommit)
            {
                if (m_NotifyOwnerOnChange)
                {
                    m_Config->m_Debouncer->Signal( ssformat( &mp, "%s|%s", m_KeyPathList[m_KeyPathCount - 1].c_str, name.c_str), 1000);
                }
                m_Config->CommitNeeded();
            }
            return key;
        }

        FacKey* CreateKey( szc_t name, int64_t timestamp)
        {
            return CreateKey( name, timestamp, true/*propagateTimestamp*/, false/*overwriteTimestamp*/, nullptr/*owningPeerId*/);
        }

        FacKey* CreateKey( szc_t name) override
        {
            mp_c mp;
            int64_t timestamp = time_100nsec();
            return CreateKey( name, timestamp);
        }

        int/*err*/ DeleteItem( szc_t name, bool leaveWhiteout, bool whiteoutOnMatchOnly, int64_t timestamp, bool propagateTimestamp, szc_t owningPeerId)
        {
            int err = err_not_found;
            bool shouldCommit = false;
            rblink_t* link;
            rblink_t* prev;
            FacWhiteout* extra = nullptr;
            FacWhiteout* whiteout = leaveWhiteout ? m_Config->WhiteoutFactory( name, timestamp) : nullptr;
            FacItem* item = nullptr;
            m_KeyLock.lock();
            link = rblink_find_2( m_Center, FacItem::Compare, name, &prev);
            if (link)
            {
                shouldCommit = true;
                item = containerof( FacItem, m_Link, link);
                rblink_remove( &m_Center, link);
                UpdateFilterChangeIdUnderLock();
                err = 0;
            }
            if (whiteout && (link || !whiteoutOnMatchOnly))
            {
                shouldCommit = true;
                rblink_insert( &m_Center, prev, &whiteout->m_Link);
                whiteout->m_Parent = this;
                Retain();
            }
            else
            {
                extra = whiteout;
            }
            m_KeyLock.unlock();
            if (item) item->Remove();
            if (extra) extra->Remove();
                // This means it changed
            if (shouldCommit && whiteout)
            {
                UpdateTimestamp( timestamp, propagateTimestamp);
                m_Config->RepChangeFactory( owningPeerId, FacItem::whiteoutType, this, name, nullptr/*val*/, timestamp);
            }
            if (shouldCommit)
            {
                m_Config->CommitNeeded();
            }
            return err;
        }

        int/*err*/ DeleteItem( szc_t name, bool whiteoutOnMatchOnly, int64_t timestamp)
        {
            return DeleteItem( name, true/*leaveWhiteout*/, whiteoutOnMatchOnly, timestamp, true/*propagateTimestamp*/, nullptr/*owningPeerId*/);
        }

        int/*err*/ /*IFacKey*/Delete( szc_t name) override
        {
            int64_t timestamp = time_100nsec();
            return DeleteItem( name, true/*whiteoutOnMatchOnly*/, timestamp);
        }

        int/*err*/ /*IFacKey*/DeleteSelf() override
        {
            int err = 0;
            if (m_Parent)
            {
                m_Parent->Delete( m_Name);
            }
            return err;
        }

        size_t/*count*/ GetKeyList( malloc_i* resMa, FacKey*** outList)
        {
            size_t keyCount = 0;
            size_t count = 0;
            FacKey** list;
            m_KeyLock.lock();
            rblink_t* link = nullptr;
            while ((link = rblink_next( m_Center, link)) != nullptr)
            {
                FacItem* item = containerof( FacItem, m_Link, link);
                if (item->m_Type == FacItem::folderType)
                {
                    keyCount++;
                }
            }
            list = ma_n_alloc( resMa, FacKey*, keyCount + 1);
            link = nullptr;
            while ((link = rblink_next( m_Center, link)) != nullptr)
            {
                FacItem* item = containerof( FacItem, m_Link, link);
                if (item->m_Type == FacItem::folderType)
                {
                    FacKey* key = SCAST( FacKey*, item);
                    key->Retain();
                    list[count] = key;
                    count++;
                }
            }
            m_KeyLock.unlock();
            if (outList) *outList = list;
            ASSERT( count == keyCount);
            return count;
        }

        size_t/*count*/ /*IFacKey*/GetValList( malloc_i* resMa, FacKeyVal*** outList) override
        {
            size_t valCount = 0;
            size_t count = 0;
            FacKeyVal** list;
            m_KeyLock.lock();
            rblink_t* link = nullptr;
            while ((link = rblink_next( m_Center, link)) != nullptr)
            {
                FacItem* item = containerof( FacItem, m_Link, link);
                if (item->m_Type == FacItem::valType)
                {
                    valCount++;
                }
            }
            list = ma_n_alloc( resMa, FacKeyVal*, valCount + 1);
            link = nullptr;
            while ((link = rblink_next( m_Center, link)) != nullptr)
            {
                FacItem* item = containerof( FacItem, m_Link, link);
                if (item->m_Type == FacItem::valType)
                {
                    FacVal* val = SCAST( FacVal*, item);
                    FacKeyVal* valKey = ma_t_alloc( resMa, FacKeyVal);
                    valKey->name = sscpy( resMa, val->m_Name);
                    valKey->val = sscpy( resMa, val->m_Val);
                    list[count] = valKey;
                    count++;
                }
            }
            m_KeyLock.unlock();
            if (outList) *outList = list;
            ASSERT( count == valCount);
            return count;
        }

        size_t/*count*/ GetKeyListOlderThan( malloc_i* resMa, IFacKey*** outList, int64_t ageNsec) override
        {
            mp_c mp;
            FacKey** keyList;
            size_t keyCount = GetKeyList( &mp, &keyList);
            size_t count = 0;
            IFacKey** list = ma_n_alloc( resMa, IFacKey*, keyCount + 1);
            size_t i;
            for (i = 0; i < keyCount; i++)
            {
                if (ageNsec == 0 || keyList[i]->m_Timestamp <= ageNsec)
                {
                    list[count] = keyList[i];
                    count++;
                }
                else
                {
                    keyList[i]->Release();
                }
            }
            if (outList) *outList = list;
            ASSERT( count <= keyCount);
            return count;
        }

        size_t/*count*/ /*IFacKey*/GetKeyList( malloc_i* resMa, IFacKey*** outList) override
        {
            return GetKeyListOlderThan( resMa, outList, 0/*ageNsec*/);
        }

        IFacList* /*IFacKey*/GetList() override
        {
            return new(t_calloc(FacList)) FacList( this);
        }

        size_t/*count*/ /*IFacKey*/GetFilteredValList( malloc_i* resMa, FacKeyVal*** outList, szc_t filterKey, szc_t filterData) override
        {
            size_t count = 0;
            FacKeyVal** list = nullptr;
            Filter* filter = FilterFactory( this, Filter::jsonFilterType, filterKey, filterData);
            if (filter)
            {
                count = filter->GetValList( resMa, &list);
                filter->Release();
            }
            if (outList) *outList = list;
            return count;
        }

        IFacList* /*IFacKey*/GetFilteredValList( szc_t filterKey, szc_t filterData) override
        {
            Filter* filter = FilterFactory( this, Filter::jsonFilterType, filterKey, filterData);
            IFacList* it = filter->GetValList();
            filter->Release();
            return it;
        }

        char* /*IFacKey*/GetKeyName( malloc_i* resMa) override
        {
            return sscpy( resMa, m_Name);
        }

        int64_t /*IFacKey*/GetKeyTimestamp() override
        {
            return m_Timestamp;
        }

        int/*err*/ SetValTimestamp( szc_t name, int64_t timestamp) override
        {
            FacVal* val = RetainVal( name);
            int err = val ? 0 : err_not_found;
            if (val)
            {
                val->m_Parent->m_KeyLock.lock();
                val->m_Timestamp = timestamp;
                val->m_Parent->m_KeyLock.unlock();

                    // Use current time here to ensure replication doesn't break
                val->m_Parent->UpdateTimestamp( time_100nsec(), true/*propagate*/);
                val->FacItem_Release();
            }
            return err;
        }

        int/*err*/ /*IFacKey*/GetValTimestamp( int64_t* outTimestamp, szc_t name) override
        {
            int64_t timestamp = 0;
            FacVal* val = RetainVal( name);
            int err = val ? 0 : err_not_found;
            if (val)
            {
                timestamp = val->m_Timestamp;
                val->FacItem_Release();
            }
            if (outTimestamp) *outTimestamp = timestamp;
            return err;
        }

        int/*err*/ GetVal( malloc_i* resMa, szc_t name, char** outstrVal)
        {
            mp_c mp;
            sz_t strVal = nullptr;
            FacVal* val = RetainVal( name);
            int err = val ? 0 : err_not_found;
            if (!err)
            {
                m_KeyLock.lock();
                strVal = val->GetVal( resMa);
                m_KeyLock.unlock();
                val->FacItem_Release();
            }
            if (outstrVal) *outstrVal = strVal;
            return err;
        }

        int/*err*/ /*IFacKey*/GetValStr( malloc_i* resMa, char** outStr, szc_t name) override
        {
            return GetVal( resMa, name, outStr);
        }

        int/*err*/ /*IFacKey*/GetValUint64( uint64_t* outUint64, szc_t name) override
        {
            mp_c mp;
            sz_t strVal;
            uint64_t valInt = 0;
            int err = GetVal( &mp, name, &strVal.c_str);
            if (!err)
            {
                valInt = sstoint( strVal, SIZE_MAX, nullptr);
            }
            *outUint64 = valInt;
            return err;
        }

        int/*err*/ /*IFacKey*/GetValInt( int* outInt, szc_t name) override
        {
            mp_c mp;
            sz_t strVal;
            int64_t valInt = 0;
            int err = GetVal( &mp, name, &strVal.c_str);
            if (!err)
            {
                valInt = sstoint( strVal, SIZE_MAX, nullptr);
            }
            SCASTASSIGN( int, *outInt, valInt);
            return err;
        }

        int/*err*/ /*IFacKey*/GetValJson( malloc_i* ma, json_t** outJson, szc_t name) override
        {
            json_t* json = nullptr;
            sz_t strVal;
            int err = GetVal( ma, name, &strVal.c_str);
            if (err) goto done;
            json = ptjson_import( ma, nullptr, strVal, sslen( strVal));
            if (!json) err = err_bad_format;
        done:
            if (outJson) *outJson = json;
            return err;
        }

        void RemoveValFromUnmatchedFiltersUnderLock( FacVal* val)
        {
            mp_c mp;
            Filter* filter;
            ASSERT( m_KeyLock.is_locked());
            rblink_t* link = nullptr;
            while ((link = rblink_next( m_CenterFilter, link)) != nullptr)
            {
                filter = containerof( Filter, m_FilterLink, link);
                if (filter->m_FilterType == val->m_FilterType)
                {
                    if (val->m_FilterType == Filter::jsonFilterType)
                    {
                        ASSERT( val->m_Json);
                        if (ss8asciicmpi( filter->m_FilterData, ptjson_get_sz(&mp, val->m_Json, filter->m_FilterKey)) != 0)
                        {
                            val->RemoveFilterLinkUnderLock( filter->m_FilterType, filter->m_FilterKey, filter->m_FilterData);
                        }
                    }
                }
            }
        }

        void AddValToMatchedFiltersUnderLock( FacVal* val)
        {
            mp_c mp;
            ASSERT( m_KeyLock.is_locked());
            rblink_t* link = nullptr;
            while ((link = rblink_next( m_CenterFilter, link)) != nullptr)
            {
                Filter* filter = containerof( Filter, m_FilterLink, link);
                if (filter->m_FilterType == val->m_FilterType)
                {
                    if (val->m_FilterType == Filter::jsonFilterType)
                    {
                        ASSERT( val->m_Json);
                        if (ss8asciicmpi( filter->m_FilterData, ptjson_get_sz(&mp, val->m_Json, filter->m_FilterKey)) == 0)
                        {
                            val->AddFilterLinkUnderLock( filter);
                        }
                    }
                }
            }
        }

        int/*err*/ SetVal( szc_t name, szc_t strVal, int64_t timestamp, bool propagateTimestamp, szc_t owningPeerId)
        {
            mp_c mp;
            int err = 0;
            int changed = 0;
            bool shouldCommit = false;
            rblink_t* collision;
            rblink_t* prev;

            FacItem* removeItem = nullptr;
            FacVal* val = FacConfig::ValFactory( name, strVal, timestamp);
            FacVal* extra = nullptr;
            if (m_Parent) m_Parent->m_KeyLock.lock();
            m_KeyLock.lock();
            if (!m_Removed)
            {
                collision = rblink_find_2( m_Center, FacItem::Compare, name, &prev);
                if (collision)
                {
                    FacItem* item = containerof( FacItem, m_Link, collision);
                    if (item->m_Type == FacItem::valType)
                    {
                        extra = val;
                        val = SCAST( FacVal*, item);
                        changed = val->SetValUnderParentLock( extra->m_Val, timestamp);
                    }
                    else
                    {
                        removeItem = item;
                        rblink_remove( &m_Center, collision);
                        collision = nullptr;
                    }
                }
                if (!collision)
                {
                    Retain();
                    val->m_Parent = this;
                    rblink_insert( &m_Center, prev, &val->m_Link);
                }
                UpdateFilterChangeIdUnderLock();
                    // FacVal changed, remove corresponding attached filter

                if (m_Parent && !err)
                {
                    if (changed)
                    {
                        RemoveValFromUnmatchedFiltersUnderLock( val);
                    }
                    if (!extra || changed)
                    {
                        AddValToMatchedFiltersUnderLock( val);
                    }
                }
                shouldCommit = true;
                val->m_Timestamp = timestamp;
            }
            else
            {
                extra = val;
                err = err_access_denied;
            }
            m_KeyLock.unlock();
            if (m_Parent) m_Parent->m_KeyLock.unlock();
            if (removeItem)
            {
                removeItem->Remove();
            }
            if (extra)
            {
                extra->Remove();
            }
            if (shouldCommit)
            {
                m_Config->CommitNeeded();
            }
            UpdateTimestamp( timestamp, propagateTimestamp);
            m_Config->RepChangeFactory( owningPeerId, FacItem::valType, this, name, strVal, timestamp);
            if (m_NotifyOwnerOnChange)
            {
                m_Config->m_Debouncer->Signal( ssformat( &mp, "%s/%s", m_KeyPathList[m_KeyPathCount - 1].c_str, name.c_str), 1000);
            }
            return err;
        }

        int/*err*/ SetVal( szc_t name, szc_t strVal, int64_t timestamp)
        {
            return SetVal( name, strVal, timestamp, true/*propagateTimestamp*/, nullptr/*owningPeerId*/);
        }

        int/*err*/ SetVal( szc_t name, szc_t strVal)
        {
            mp_c mp;
            int64_t timestamp = time_100nsec();
            return SetVal( name, strVal, timestamp);
        }

        int/*err*/ /*IFacKey*/SetValStr( szc_t name, szc_t data) override
        {
            return SetVal( name, data);
        }

        int/*err*/ /*IFacKey*/SetValInt( szc_t name, int data) override
        {
            mp_c mp;
            return SetVal( name, ssformat( &mp, "%i", data));
        }

        int/*err*/ /*IFacKey*/SetValUint64( szc_t name, uint64_t data) override
        {
            mp_c mp;
            return SetVal( name, ssformat( &mp, "%I64u", data));
        }

        int/*err*/ /*IFacKey*/SetValJson( szc_t name, json_t* json) override
        {
            mp_c mp;
            sz_t sz = ptjson_export( &mp, nullptr, json, false);
            ASSERT( sz);
            return SetVal( name, sz);
        }

        int/*err*/ /*IFacKey*/SetKeyInJson( szc_t name, szc_t jsonKey, json_t* jsonValue) override
        {
            int err = 0;
            mp_c mp;
            m_JsonKeyMutex.lock();
            json_t* json;
            GetValJson( &mp, &json, name);
            if (!json && jsonValue) json = ptjson_new_object( &mp);
            if (json)
            {
                    // Set to nullptr is alias for delete
                if (!jsonValue) ptjson_remove( &mp, json, jsonKey);
                else ptjson_set_object( &mp, json, jsonKey, jsonValue);
                err = SetVal( name, ptjson_export( &mp, nullptr, json, false));
            }
            m_JsonKeyMutex.unlock();
            return err;
        }

        size_t/*count*/ /*IFacKey*/GetValNameListOlderThan( malloc_i* resMa, sz_t** outValNameList, int64_t ageNsec) override
        {
            size_t count = 0;
            size_t valCount = 0;
            FacItem* item;
            FacVal* val;
            sz_t* valNameList;
            m_KeyLock.lock();
            rblink_t* link = nullptr;
            while ((link = rblink_next( m_Center, link)) != nullptr)
            {
                valCount++;
            }
            valNameList = ma_n_alloc( resMa, sz_t, valCount + 1);
            link = nullptr;
            while ((link = rblink_next( m_Center, link)) != nullptr)
            {
                item = containerof( FacItem, m_Link, link);
                if (item->m_Type == FacItem::valType &&
                    (ageNsec == 0 || item->m_Timestamp <= ageNsec))
                {
                    val = SCAST( FacVal*, item);
                    valNameList[count] = sscpy( resMa, val->m_Name);
                    count++;
                }
            }
            m_KeyLock.unlock();
            *outValNameList = valNameList;
            ASSERT( count <= valCount);
            return count;
        }

        size_t/*count*/ /*IFacKey*/GetValNameList( malloc_i* resMa, sz_t** outValNameList) override
        {
            return GetValNameListOlderThan( resMa, outValNameList, 0/*ageNsec*/);
        }

        void /*IFacKey*/ToggleNotify( bool notifyOnChildChange) override
        {
            m_NotifyOwnerOnChange = notifyOnChildChange;
        }

        void /*IFacKey*/Empty() override
        {
            FacItem* item;
            Filter* filter;
            rblink_t* link;

            m_KeyLock.lock();
            while ((link = m_CenterFilter) != nullptr)
            {
                rblink_remove( &m_CenterFilter, link);
                m_KeyLock.unlock();
                filter = containerof( Filter, m_FilterLink, link);
                filter->Release();
                m_KeyLock.lock();
            }
            while ((link = m_Center) != nullptr)
            {
                UpdateFilterChangeIdUnderLock();
                rblink_remove( &m_Center, link);
                m_KeyLock.unlock();
                item = containerof( FacItem, m_Link, link);
                item->Remove();
                m_KeyLock.lock();
            }
            m_KeyLock.unlock();
        }

        void LoadJson( json_t* json, bool propogateTimestamps)
        {
            mp_c mp;
            bool isArray = false;
            FacKey* newKey;
            size_t keysLoaded = 0;
            size_t valsLoaded = 0;
            size_t whiteoutsLoaded = 0;
            size_t count = 0;
            size_t i;
            const char *keySz;
            szc_t strVal = nullptr;
            json_t* val = nullptr;
            char** keys = nullptr;
            json_t** vals = nullptr;
            Empty();
            if (ptjson_is_object( json))
            {
                count = ptjson_get_list( &mp, &keys, &vals, json);
            }
            else if (ptjson_is_array( json))
            {
                isArray = true;
                count = ptjson_get_array_count( json);
            }
            else
            {
                ASSERT( false && "Unknown json type for LoadJson");
            }

            for (i = 0; i < count; i++)
            {
                if (!isArray)
                {
                    keySz = keys[i];
                    val = vals[i];
                }
                else
                {
                    keySz = ssformat( &mp, "%i", i);
                    val = ptjson_get_array_member( json, i);
                }
                if (val)
                {
                    bool valIsArray = ptjson_is_array( val);
                    if (valIsArray || ptjson_is_object( val))
                    {
                        int64_t timestamp = 0;
                        bool isVal = false;
                        bool isWhiteout = false;
                        if (!valIsArray)
                        {
                            timestamp = SzToTimestamp( ptjson_get_sz( &mp, val, "$t"));
                            json_t* valField = ptjson_get_object( val, "$v");
                            json_t* whiteoutField = ptjson_get_object( val, "$w");
                            if (valField)
                            {
                                isVal = true;
                                valsLoaded++;
                                SetVal( keySz, ptjson_as_sz( &mp, valField), timestamp, propogateTimestamps, nullptr/*owningPeerId*/);
                            }
                            else if (whiteoutField)
                            {
                                isWhiteout = true;
                                whiteoutsLoaded++;
                                DeleteItem( keySz, true/*leaveWhiteout*/, false/*whiteoutOnMatchOnly*/, timestamp, propogateTimestamps, nullptr/*owningPeerId*/);
                            }
                        }
                        if (!isVal && !isWhiteout)
                        {
                            newKey = CreateKey( keySz, timestamp, propogateTimestamps, false/*overwriteTimestamp*/, nullptr/*owningPeerId*/);
                            newKey->LoadJson( val, propogateTimestamps);
                            keysLoaded++;
                            newKey->Release();
                        }
                    }
                    else
                    {
                            // Legacy v7 config, no timestamps, no whiteouts
                        uint64_t intVal = 0;
                        strVal = nullptr;
                        if (ptjson_is_string( val))
                        {
                            strVal = ptjson_as_sz( &mp, val);
                        }
                        else if (ptjson_is_int( val))
                        {
                            intVal = ptjson_as_int( val);
                            strVal = ssformat( &mp, "%I64u", intVal);
                        }
                        else
                        {
                            ASSERT( ssformat( &mp, "Unsupported json type: %s", keySz));
                        }
                        ASSERT (ss8asciicmpi( keySz, "$v") != 0);
                        ASSERT (ss8asciicmpi( keySz, "$w") != 0);
                        if (ss8asciicmpi( keySz, "$t") == 0)
                        {
                            m_Timestamp = SzToTimestamp( strVal);
                        }
                        else
                        {
                            if (isArray)
                            {
                                keySz = strVal;
                            }
                            SetVal( keySz, strVal);
                            valsLoaded++;
                        }
                    }
                }
            }
        }

        void LoadJson( json_t* json) override
        {
            LoadJson( json, true/*propogateTimestamps*/);
        }


        static json_t* ConvertToJsonObject( malloc_i* resMa, FacKey* key)
        {
            mp_c mp;
            UidSet deleteAfterWithNoWhiteouts;
            rblink_t* link = nullptr;
            FacItem* subItem;
            FacKey* subKey;
            FacWhiteout* subWhiteout;
            json_t* jsonCont = ptjson_new_object( resMa);
            FacVal* val;
            key->m_KeyLock.lock();
            ptjson_set_sz( resMa, jsonCont, "$t", TimestampToSz( resMa, key->m_Timestamp));
            while ((link = rblink_next( key->m_Center, link)) != nullptr)
            {
                subItem = containerof( FacItem, m_Link, link);
                if (subItem->m_Type == FacItem::folderType)
                {
                    subKey = SCAST( FacKey*, subItem);
                    json_t* json = ConvertToJsonObject( resMa, subKey);
                    ptjson_set_object( resMa, jsonCont, subKey->m_Name, json);
                }
                else if (subItem->m_Type == FacItem::valType)
                {
                    val = SCAST( FacVal*, subItem);
                    ASSERT( ss8asciicmpi( subItem->m_Name, "$t") != 0);
                    ASSERT( ss8asciicmpi( subItem->m_Name, "$v") != 0);
                    json_t* cont = ptjson_setv( resMa, ptjson_new_object( resMa),
                        pjv( "$t", TimestampToSz( &mp, val->m_Timestamp)),
                        pjv( "$v", val->m_Val),
                    pjv_end());
                    ptjson_set_object( resMa, jsonCont, val->m_Name, cont);
                }
                else if (subItem->m_Type == FacItem::whiteoutType)
                {
                    subWhiteout = SCAST( FacWhiteout*, subItem);
                        // Clean up expired whiteouts on saves
                    if (!FacWhiteout::Expired( subWhiteout->m_Timestamp))
                    {
                        json_t* cont = ptjson_setv( resMa, ptjson_new_object( resMa),
                            pjv( "$t", TimestampToSz( resMa, subWhiteout->m_Timestamp)),
                            pjv( "$w", 1),
                        pjv_end());
                        ptjson_set_object( resMa, jsonCont, subWhiteout->m_Name, cont);
                    }
                    else
                    {
                        deleteAfterWithNoWhiteouts.Add( subWhiteout->m_Name);
                    }
                }
            }
            key->m_KeyLock.unlock();
            sz_t* list;
            size_t i;
            size_t count = deleteAfterWithNoWhiteouts.GetList( &mp, &list);
            for (i = 0; i < count; i++)
            {
                    // Clean up expired whiteouts
                key->m_Config->tracef( "Cleaning up expired whiteout %s", list[i].c_str);
                key->DeleteItem( list[i], false/*leaveWhiteout*/, false/*whiteoutOnMatchOnly*/, 0, false/*propagateTimestamp*/, nullptr/*owningPeerId*/);
            }
            return jsonCont;
        }

        char* /*IFacKey*/ConvertToJson( malloc_i* resMa) override
        {
            mp_c mp;
            char* jsonBlob;
            json_t* json = ConvertToJsonObject( &mp, this);
            jsonBlob = ptjson_export( resMa, nullptr, json, false);

            return jsonBlob;
        }
    };

    FacKey* KeyFactory( szc_t name)
    {
        return new(t_calloc(FacKey)) FacKey( this, name);
    }

    void /*IFacConfig*/EmptyConfig() override
    {
        root->Empty();
    }

    void BackupFile( szc_t fileLocation)
    {
        mp_c mp;
        FD_T fd;
        FD_T fd2;
        uint64_t size;
        char* buffer;
        size_t actualSize;
        szc_t newLocation;
        int err = f8_open( &fd, fileLocation, fof_read);
        if (err) goto done;

        file_get_size( fd, &size);
        buffer = ma_n_alloc( &mp, char, size + 1);
        file_set_offset( fd, 0);
        err = file_read( fd, buffer, size, &actualSize);
        if (err) goto done;

        int year, month, day, hour, minute, second;
        file_time_unpack( time_100nsec(), &year, &month, &day, &hour, &minute, &second, nullptr);
        newLocation = ssformat( &mp, "%s.%04u%02u%02u_%02u%02u%02u.bak", fileLocation.c_str, year, month, day, hour, minute, second);
        f8_delete( newLocation);
        err = f8_open( &fd2, newLocation, fof_create | fof_write);
        if (!err)
        {
            file_write( fd2, buffer, actualSize, nullptr);
            file_set_size( fd2, actualSize);
            file_close( fd2);
        }
        file_close( fd);
    done:
        return;
    }

    int/*err*/ Save()
    {
        mp_c mp;
        json_t* json = FacKey::ConvertToJsonObject( &mp, root);
        size_t configSize;
        sz_t configData = ptjson_export( &mp, &configSize, json, true);
        FD_T fd;
        int err = f8_open(&fd, m_StorageLocation, fof_create | fof_write);
        if (!err)
        {
            file_write(fd, configData, configSize, nullptr);
            file_set_size(fd, configSize);
            fd_close(fd);
        }
        return err;
    }

    int/*err*/ /*IFacConfig*/Commit() override
    {
        return Save();
    }

    int/*err*/ /*IFacConfig*/SaveBackup() override
    {
        BackupFile( m_StorageLocation);
        return 0;
    }

    int/*err*/ LoadConfig()
    {
        mp_c mp;
        if (m_Output) m_Output->outf( "Loading config: %s\n", m_StorageLocation.c_str);
        int err = 0;
        json_t* config;
        err = FileReadJson( &mp, &config, m_StorageLocation);
        if (config)
        {
            root->LoadJson( config, false/*propogateTimestamps*/);
{
TRAP;
json_t* json2 = FacKey::ConvertToJsonObject( &mp, root);
size_t configSize;
sz_t configData = ptjson_export( &mp, &configSize, json2, true);
FD_T fd2;
f8_open( &fd2, sscatm( &mp, m_StorageLocation, ".joe"), fof_create | fof_write);
file_write( fd2, configData, configSize, nullptr);
file_set_size( fd2, configSize);
fd_close( fd2);
}
        }
        else if (f8_exists( m_StorageLocation, nullptr) == 0)
        {
            m_FailedLoad = true;
        }
        return err;
    }

    IFacKey* OpenKey( szc_t name) override { return root->OpenKey( name); }
    IFacKey* CreateKey( szc_t name) override { return root->CreateKey( name); }
    int/*err*/ Delete( szc_t name) override { return root->Delete( name); }
    int/*err*/ DeleteSelf() override { return root->DeleteSelf(); }
    size_t/*count*/ GetKeyList( malloc_i* resMa, IFacKey*** outList) override { return root->GetKeyList( resMa, outList); }
    size_t/*count*/ GetKeyListOlderThan( malloc_i* resMa, IFacKey*** outList, int64_t ageNsec) override { return root->GetKeyListOlderThan( resMa, outList, ageNsec); }
    IFacList* GetList() override { return root->GetList(); }
    size_t/*count*/ GetValList( malloc_i* resMa, FacKeyVal*** outList) override { return root->GetValList( resMa, outList); }
    size_t/*count*/ GetFilteredValList( malloc_i* resMa, FacKeyVal*** outList, szc_t filterKey, szc_t filterData) override { return root->GetFilteredValList( resMa, outList, filterKey, filterData); }
    IFacList* GetFilteredValList( szc_t valName, szc_t data) override { return root->GetFilteredValList( valName, data); }
    char* GetKeyName( malloc_i* resMa) override { return root->GetKeyName( resMa); }
    int64_t GetKeyTimestamp() override { return root->GetKeyTimestamp(); }
    int SetValTimestamp( szc_t name, int64_t timestamp) override { return root->SetValTimestamp( name, timestamp); }
    int GetValTimestamp( int64_t* outTimestamp, szc_t name) override { return root->GetValTimestamp( outTimestamp, name); }
    int GetValStr( malloc_i* resMa, char** outStr, szc_t name) override { return root->GetValStr( resMa, outStr, name); }
    int GetValInt( int* outInt, szc_t name) override { return root->GetValInt( outInt, name); }
    int GetValUint64( uint64_t* outUint64, szc_t name) override { return root->GetValUint64( outUint64, name); }
    int GetValJson( malloc_i* ma, json_t** outJson, szc_t name) override { return root->GetValJson( ma, outJson, name); }
    int SetValStr( szc_t name, szc_t data) override { return root->SetValStr( name, data); }
    int SetValInt( szc_t name, int data) override { return root->SetValInt( name, data); }
    int SetValUint64( szc_t name, uint64_t data) override { return root->SetValUint64( name, data); }
    int SetValJson( szc_t name, json_t* json) override { return root->SetValJson( name, json); }
    int SetKeyInJson( szc_t name, szc_t jsonKey, json_t* jsonValue) override { return root->SetKeyInJson( name, jsonKey, jsonValue); }
    char* ConvertToJson( malloc_i* resMa) override { return root->ConvertToJson( resMa); }
    size_t/*count*/ GetValNameList( malloc_i* resMa, sz_t** outValNameList) override { return root->GetValNameList( resMa, outValNameList); }
    size_t/*count*/ GetValNameListOlderThan( malloc_i* resMa, sz_t** outValNameList, int64_t ageNsec) override { return root->GetValNameListOlderThan( resMa, outValNameList, ageNsec); }
    void LoadJson( json_t* json) override { root->LoadJson( json); }
    void ToggleNotify( bool shouldNotify) override { root->ToggleNotify( shouldNotify); }
    void Empty() override { root->Empty(); }

    static json_t* PackItemNode( malloc_i* ma, const ItemNode* node)
    {
        if (!node) return nullptr;
        json_t* json = ptjson_setv( ma, ptjson_new_object( ma),
            pjv( "name"     , node->name     ),
            pjv( "timestamp", node->timestamp),
            pjv( "type"     , node->type     ),
            pjv_end());

            // We only fill this when a val is set
        if (!node->val.isempty()) ptjson_set_sz( ma, json, "val", node->val);
        return json;
    }

    static ItemNode* UnpackItemNode( malloc_i* resMa, json_t* json)
    {
        ItemNode* node = ma_t_alloc( resMa, ItemNode);
        ptjson_getv(resMa, json,
            pjv( "name"     , &node->name     ),
            pjv( "val"      , &node->val      ),
            pjv( "timestamp", &node->timestamp),
            pjv( "type"     , &node->type     ),
            pjv_end());
        return node;
    }

    struct Channel;
    struct Peer
    {
        // pin_wire_t m_CancelWire;
        rblink_t m_PeerLink;
        // pin_c m_CancelPin;
        FacConfig* m_Config;
        sz_c m_PeerId;
        sz_c m_PeerPriority; // Used to establish slave/master relationship
        int m_RefCount;
        int m_Relationship;
        bool m_Cancelled;
        Channel* m_Channel;
        sz_c m_ServerName;
        int m_ServerRole;

        struct Relationship
        {
            static const int Slave  = 1;
            static const int Master = 2;
            static const int None   = 3;
        };

            // This constructor expects zero-initialized memory.
        void* operator new( size_t size, void* mem) { return mem; }
        Peer( FacConfig* config, szc_t peerId, szc_t peerPriority, szc_t serverName)
        :
            m_Config( config),
            m_PeerId( peerId),
            m_PeerPriority( peerPriority),
            m_ServerName( serverName),
            m_Relationship( Relationship::None),
            m_RefCount( 1)
        {
            m_Config->BusyRetain();
        }

        ~Peer()
        {
            ASSERT( !m_RefCount);
            ASSERT( !m_Channel);
            m_Config->BusyRelease();
        }

        void ReleaseChannel( Channel* channel)
        {
            bool removed = false;
            m_Config->m_MainLock.lock();
            if (m_Channel == channel)
            {
                m_Channel = nullptr;
                removed = true;
            }
            m_Config->m_MainLock.unlock();
            if (removed)
            {
                channel->LocalActiveRelease();
            }
        }

        void SetActiveChannel( Channel* channel)
        {
            bool replicationNeeded = false;
            Channel* oldChannel = nullptr;
            m_Config->m_MainLock.lock();
            if (!m_Cancelled && (!m_Channel || memcmp( m_Channel->m_ChannelPriority, channel->m_ChannelPriority, 32) < 0))
            {
                oldChannel = m_Channel;
                if (channel) channel->LocalActiveRetain();
                m_Channel = channel;
                replicationNeeded = true;
            }
            m_Config->m_MainLock.unlock();
            if (replicationNeeded)
            {
                m_Config->QueueReplication();
            }
            if (oldChannel) oldChannel->LocalActiveRelease();
        }

        void Cancel()
        {
            m_Cancelled = true;
            bool priorityChanged = false;
            m_Config->m_MainLock.lock();
            if (rblink_is_linked( m_Config->m_CenterPeer, &m_PeerLink))
            {
                rblink_remove( &m_Config->m_CenterPeer, &m_PeerLink);
            }
            if (m_Config->m_MasterPeer == this)
            {
                m_Config->m_MasterPeer = nullptr;
                priorityChanged = true;
            }
            Channel* channel = m_Channel;
            m_Channel = nullptr;
            m_Config->m_MainLock.unlock();
            if (priorityChanged) m_Config->QueueReplication();
            if (channel)
            {
                channel->LocalActiveRelease();
            }
        }

        void Retain()
        {
            ASSERT( m_RefCount >= 0);
            sync_inc( &m_RefCount);
        }

        void Release()
        {
            ASSERT( m_RefCount > 0);
            int count = m_RefCount - 1;
            sync_compiler();
            if (count == 0 || !sync_xchg( &m_RefCount, count + 1, count))
            {
                m_Config->m_MainLock.lock();
                count = sync_dec( &m_RefCount);
                if (count == 0)
                {
                    if (rblink_is_linked( m_Config->m_CenterPeer, &m_PeerLink))
                    {
                        rblink_remove( &m_Config->m_CenterPeer, &m_PeerLink);
                    }
                }
                m_Config->m_MainLock.unlock();
                if (count == 0)
                {
                    Cancel();
                    this->~Peer();
                    free( this);
                }
            }
        }

        int GetRelationship()
        {
            m_Config->m_MainLock.lock();
            int relationship = m_Relationship;
            m_Config->m_MainLock.unlock();
            return relationship;
        }

        void SetPriority( szc_t priority)
        {
            bool priorityChanged = false;
            m_Config->m_MainLock.lock();
            if (sscmp( m_PeerPriority, priority) != 0)
            {
                m_Config->tracef( "Changed priority from %s to %s", m_PeerPriority.c_str, priority.c_str);
                m_PeerPriority = priority;
                priorityChanged = true;
            }
            m_Config->m_MainLock.unlock();
            if (priorityChanged) m_Config->QueueReplication();
        }

        Channel* RetainActiveChannel()
        {
            m_Config->m_MainLock.lock();
            Channel* channel = m_Channel;
            if (channel)
            {
                ASSERT( channel->m_Peer == this);
                channel->LocalActiveRetain();
            }
            m_Config->m_MainLock.unlock();
            return channel;
        }

        static int Compare(void* peerId1_, rblink_t* peerLink_)
        {
            szc_t peerId1 = SCAST( const char*, peerId1_);
            Peer* peer2 = containerof( Peer, m_PeerLink, peerLink_);
            return sscmp( peerId1, peer2->m_PeerId);
        }

        int/*err*/ SendJsonV( malloc_i* ma, pin_i* cancelPin, szc_t methodName, ...)
        {
            Channel* channel = RetainActiveChannel();
            va_list va;
            va_start( va, methodName);
            remote_i* remote = channel ? channel->m_Remote : nullptr;
            int err = FacRemoteSendJsonVa( va, remote, ma, cancelPin, Channel::m_SendTimeout, methodName);
            va_end( va);
            if (channel) channel->LocalActiveRelease();
            return err;
        }

        struct PeerRequestContext
        {
            Channel* m_Channel;
            FacRequestComplete m_Complete;
            void* m_Context;

            PeerRequestContext( Peer* peer, FacRequestComplete complete, void* context)
            :
                m_Channel( peer->RetainActiveChannel()),
                m_Complete( complete),
                m_Context( context)
            {
            }

            ~PeerRequestContext()
            {
                if (m_Channel) m_Channel->LocalActiveRelease();
            }

            static void Complete( void* rx_, int inErr, int result, const void* output, size_t outputSize, FD_T outputFds[ remote_max_fd_count])
            {
                PeerRequestContext* rx = SCAST( PeerRequestContext*, rx_);
                int err = inErr;
                if (!err) err = perr_to_err( result);
                if (rx->m_Complete) rx->m_Complete( rx->m_Context, err, RCAST( const char*, output), outputSize);
                delete rx;
            }
        };

        void SendProxyRequest( pin_i* cancelPin, FacRequestComplete complete, void* context, szc_t methodName, szc_t input)
        {
            mp_c mp;
            json_t* input2j = ptjson_setv( &mp, ptjson_new_object( &mp),
                pjv( "method"    , methodName),
                pjv_enter( "args"),
                    pjv( "request", ptjson_import( &mp, nullptr, input, sslen( input))),
                pjv_leave(),
                pjv_end());
            sz_t input2 = ptjson_export( &mp, nullptr, input2j, false);
            PeerRequestContext* rx = new PeerRequestContext( this, complete, context);
            if (rx->m_Channel)
            {
                rx->m_Channel->m_Remote->send_async( cancelPin, -1, FacWire::MethodCode::Json, input2, sslen( input2), nullptr, 10000000, PeerRequestContext::Complete, rx);
            }
            else
            {
                rx->Complete( rx, err_broken_pipe, 0, nullptr, 0, nullptr);
            }
        }

        int/*err*/ SetVal( sz_t* keyPathList, size_t keyPathCount, szc_t name, szc_t val, int64_t timestamp, bool propagateTimestamp)
        {
            mp_c mp;
            int err = SendJsonV( &mp, nullptr, "set_val",
                pjv( "key_path_list"      , PackStringList( &mp, keyPathList, keyPathCount)),
                pjv( "name"               , name              ),
                pjv( "val"                , val               ),
                pjv( "timestamp"          , timestamp         ),
                pjv( "propagate_timestamp", propagateTimestamp),
                pjv_end(),
                pjv_end());
            return err;
        }

        int/*err*/ CreateKey( sz_t* keyPathList, size_t keyPathCount, int64_t timestamp, bool propagateTimestamp, bool overwriteTimestamp)
        {
            mp_c mp;
            int err = SendJsonV( &mp, nullptr, "create_key",
                pjv( "key_path_list"      , PackStringList( &mp, keyPathList, keyPathCount)),
                pjv( "timestamp"          , timestamp         ),
                pjv( "propagate_timestamp", propagateTimestamp),
                pjv( "overwrite_timestamp", overwriteTimestamp),
                pjv_end(),
                pjv_end());
            return err;
        }

        int/*err*/ DeleteItem( sz_t* keyPathList, size_t keyPathCount, szc_t name, bool leaveWhiteout, int64_t timestamp, bool propagateTimestamp)
        {
            mp_c mp;
            int err = SendJsonV( &mp, nullptr, "delete_item",
                pjv( "key_path_list"      , PackStringList( &mp, keyPathList, keyPathCount)),
                pjv( "name"               , name              ),
                pjv( "leave_whiteout"     , leaveWhiteout     ),
                pjv( "timestamp"          , timestamp         ),
                pjv( "propagate_timestamp", propagateTimestamp),
                pjv_end(),
                pjv_end());
            return err;
        }

        int/*err*/ SubscribeTo( malloc_i* ma, sz_t* outMasterPriority, bool toggle)
        {
            ASSERT( ma);
            sz_t masterPriority;
            int err = SendJsonV( ma, nullptr, "subscribe_to_master",
                pjv( "toggle"         , toggle         ),
                pjv_end(),
                pjv( "master_priority", &masterPriority),
                pjv_end());
                // Will queue electmaster on change
            if (!masterPriority.isempty()) SetPriority( masterPriority);
            if (outMasterPriority) *outMasterPriority = sscpy( ma, masterPriority);
            return err;
        }

        size_t/*count*/ ListItems( malloc_i* ma, ItemNode*** outList, sz_t* keyPathList, size_t keyPathCount)
        {
            ASSERT( ma);
            json_t* jsonList;
            int check = SendJsonV( ma, nullptr, "list_items",
                pjv( "key_path_list", PackStringList( ma, keyPathList, keyPathCount)),
                pjv_end(),
                pjv( "list"         , &jsonList),
                pjv_end());
            size_t count = 0;
            ItemNode** list = ma_n_alloc( ma, ItemNode*, 1);
            if (!check)
            {
                count = ptjson_get_array_count( jsonList);
                list = ma_n_alloc( ma, ItemNode*, count + 1);
                size_t i;
                for (i = 0; i < count; i ++)
                {
                    list[i] = UnpackItemNode( ma, ptjson_get_array_member( jsonList, i));
                }
            }
            if (outList) *outList = list;
            return count;
        }

        void UpdatePeerPriority( szc_t priority)
        {
            TRACE_ENTER();
            mp_c mp;
            SendJsonV( &mp, nullptr, "update_peer_priority",
                pjv( "priority", priority),
                pjv_end(),
                pjv_end());
            TRACE_EXIT();
        }

        int/*err*/ RelayHub( malloc_i* ma, json_t** outJsonResults, json_t* args, size_t hopCount)
        {
            json_t* jsonResults;
            int err = SendJsonV( ma, nullptr, "relay_hub",
                pjv( "args"     , args        ),
                pjv( "hop_count", hopCount    ),
                pjv_end(),
                pjv( "results"  , &jsonResults),
                pjv_end());
            *outJsonResults = jsonResults;
            return err;
        }
    };

    int/*err*/ RelayHub( malloc_i* resMa, json_t** outResultsObj, szc_t ignorePeerId, json_t* args, size_t hopCount)
    {
        FacAcx acx( &m_CancelPin);
        malloc_i* ma = acx.GetMa();
        static const size_t MAX_HOP_COUNT = 10;
        json_t* jsonObj = ptjson_new_object( resMa);
        json_t* json;
        Peer** peerList;
        json_t* jsonResponse = nullptr;
        size_t outputSize;
        sz_t output;
        sz_t input = ptjson_export( ma, nullptr, args, false);
        int err = HubAPIProcessRequest( &acx, m_Hub, input, &output, &outputSize);
        if (!err)
        {
            jsonResponse = ptjson_import( resMa, nullptr, SCAST( const char*, output), outputSize);
        }
        if (jsonResponse)
        {
            m_MainLock.lock();
            szc_t id = sscpy( ma, m_LocalId);
            m_MainLock.unlock();
            ptjson_set_object( resMa, jsonObj, id, jsonResponse);
        }

        if (hopCount < MAX_HOP_COUNT)
        {
            size_t peerCount = RetainPeerList( ma, &peerList);
            size_t i;
            for (i = 0; i < peerCount; i++)
            {
                Peer* peer = peerList[i];
                int relationship = peer->GetRelationship();
if (sscmp( peer->m_PeerId, ignorePeerId) != 0) tracef( "Relay hub check, peer %s, relationship %d", peer->m_PeerId.c_str, relationship);
                    // Skip the peer that sent the request
                if (sscmp( peer->m_PeerId, ignorePeerId) != 0 &&
                    relationship != Peer::Relationship::None)
                {
tracef( "Relay hub repeater to peer %s", peer->m_PeerId.c_str);
                    int check = peer->RelayHub( resMa, &json, args, hopCount + 1);
                    if (!check)
                    {
                        char** keys;
                        json_t** vals;
                        size_t keyCount = ptjson_get_list( ma, &keys, &vals, json);
                        size_t j;
                        for (j = 0; j < keyCount; j++)
                        {
tracef( "Relay hub added remote info from peer %s", keys[j]);
                            ptjson_set_object( resMa, jsonObj, keys[j], vals[j]);
                        }
                    }
                }
                peer->Release();
            }
        }
        *outResultsObj = jsonObj;
        return err;
    }

    struct RepChange
    {
        sz_c owningPeerId;
        int type; // Folder or val
        mp_c keyPathMp;
        sz_t* keyPathList;
        size_t keyPathCount;
        sz_c name;
        sz_c val;
        int64_t timestamp;
        RepChange* next;

            // This constructor expects zero-initialized memory.
        void* operator new(size_t size, void* mem) { return mem; }
        RepChange()
        {
        }

        ~RepChange()
        {
        }

        void Release()
        {
            this->~RepChange();
            free( this);
        }
    };

    void InsertRepChange( RepChange* repChange)
    {
        m_MainLock.lock();
        if (!m_RepChangeListFirst)
        {
            m_RepChangeListFirst = repChange;
        }
        if (m_RepChangeListLast)
        {
            m_RepChangeListLast->next = repChange;
        }
        m_RepChangeListLast = repChange;
        m_MainLock.unlock();
    }

    void RepChangeFactory( szc_t owningPeerId, int type, FacKey* parentKey, szc_t name, szc_t val, int64_t timestamp)
    {
        if (m_Started)
        {
            RepChange* repChange = new(t_calloc(RepChange)) RepChange();
            repChange->owningPeerId = owningPeerId;
            repChange->type = type;
            repChange->keyPathList = CopySzList( &repChange->keyPathMp, parentKey->m_KeyPathList, parentKey->m_KeyPathCount);
            repChange->keyPathCount = parentKey->m_KeyPathCount;
            repChange->name = name;
            repChange->val = val;
            repChange->timestamp = timestamp;
            InsertRepChange( repChange);
            ReplicationNeeded();
        }
    }

    RepChange* PopRepChange()
    {
        RepChange* repChange = nullptr;
        m_MainLock.lock();
        if (m_RepChangeListLast == m_RepChangeListFirst)
        {
            m_RepChangeListLast = nullptr;
        }
        repChange = m_RepChangeListFirst;
        if (m_RepChangeListFirst)
        {
            m_RepChangeListFirst = m_RepChangeListFirst->next;
            repChange->next = nullptr;
        }
        m_MainLock.unlock();
        return repChange;
    }

    struct Channel
    :
        IFacConfigChannel,
        remote_receive_i
    {
        pin_wire_t m_CancelWire;
        alarm_t m_ConnectAlarmWire;
        remote_i* m_Remote;
        int m_BusyCount;
        int m_ActiveCount;
        int m_LocalActiveCount;
        int/*bool*/ m_RemoteIsActive;
        uint8_t m_ChannelPriority[32];
        uint8_t m_HalfChannelPriority[32];
        FacConfig* m_Config;
        Peer* m_Peer;
        sz_c m_PeerId;
        sz_c m_ServerName;
        int m_ServerRole;
        sz_c m_PeerPriority;
        bool m_ChannelInitiator;
        int m_IdentifyPending;
        int m_IdentifyingCount;
        pin_c m_CancelPin;
        sz_c m_RemoteAddress;
        sz_c m_LocalAddress;

#ifdef DEBUG
        static const int m_ConnectTimeout = 60000;
        static const int m_SendTimeout = 60000;
#else
        static const int m_ConnectTimeout = 5000;
        static const int m_SendTimeout = 5000;
#endif
        static const size_t m_MaxOutputSize = 0x1000000;
        struct Wire
        {
            static const int MethodIdentify        = 1;
            static const int Json                  = 2/*FacWire::Json*/;
            static const int MethodCloseIfInactive = 3;
        };

        Channel( FacConfig* config, bool channelInitiator, szc_t remoteAddress, szc_t localAddress)
        :
            m_Config(config),
            m_BusyCount( 1),
            m_ActiveCount( 2),
            m_LocalActiveCount( 1),
            m_RemoteIsActive( true),
            m_ChannelInitiator( channelInitiator),
            m_RemoteAddress( remoteAddress),
            m_LocalAddress( localAddress)
        {
            m_Config->Retain();
        }

        ~Channel()
        {
        }

        void Final()
        {
            if (m_Config) m_Config->Release();
            if (m_Peer) m_Peer->Release();
            if (m_Remote) m_Remote->release();
            this->~Channel();
            free( this);
        }

        static void FinalThreadMain( void* ch_)
        {
            Channel* ch = SCAST( Channel*, ch_);
            ch->m_Peer->m_Config->m_Owner->PeerDisconnected( ch->m_PeerId);
            ch->Final();
        }

        void BusyRetain()
        {
            ASSERT( m_BusyCount > 0);
            sync_inc( &m_BusyCount);
        }

        void BusyRelease()
        {
            ASSERT( m_BusyCount > 0);
            int count = sync_dec( &m_BusyCount);
            if (count == 0)
            {
                ASSERT( m_ActiveCount == 0);
                ASSERT( m_LocalActiveCount == 0);
                ASSERT( !m_RemoteIsActive);
                if (m_Config && m_Peer && m_PeerId)
                {
                    thread_start_fast( FinalThreadMain, this);
                }
                else
                {
                    Final();
                }
            }
        }

        void Cancel()
        {
            m_CancelPin.signal();
            if (pin_i_detach( &m_Config->m_CancelPin, &m_CancelWire))
            {
                BusyRelease();
            }
            if (alarm_cancel( &m_ConnectAlarmWire))
            {
                BusyRelease();
            }
            if (m_Remote) m_Remote->cancel();
            if (sync_xchg( &m_RemoteIsActive, true, false))
            {
                ActiveRelease();
            }
            if (sync_xchg( &m_IdentifyPending, true, false))
            {
                if (sync_dec( &m_IdentifyingCount) == 0)
                {
                    LocalActiveRelease();
                }
            }
                // m_Peer is unset when the channel is lower priority than the current channel
            if (m_Peer) m_Peer->ReleaseChannel( this);
        }

        static void CCALL CancelHandler( pin_wire_t* th_)
        {
            Channel* th = containerof( Channel, m_CancelWire, th_);
            th->Cancel();
            th->BusyRelease();
        }

        static void CCALL ConnectAlarmHandler( alarm_t* th_)
        {
            Channel* th = containerof( Channel, m_ConnectAlarmWire, th_);
            th->Cancel();
            th->BusyRelease();
        }

        // void ActiveRetain()
        // {
        //     ASSERT( m_ActiveCount > 0);
        //     sync_inc( &m_ActiveCount);
        // }

        void ActiveRelease()
        {
            ASSERT( m_ActiveCount > 0);
            if (sync_dec( &m_ActiveCount) == 0)
            {
                ASSERT( m_LocalActiveCount == 0);
                ASSERT( !m_RemoteIsActive);
                Cancel();
                BusyRelease();
            }
        }

        void LocalActiveRetain()
        {
            ASSERT( m_LocalActiveCount > 0);
            sync_inc( &m_LocalActiveCount);
        }

        static void CCALL RemoteCloseComplete( void* th_, int err, int res, const void* output, size_t outputSize, FD_T unused1[])
        {
            Channel* th = SCAST( Channel*, th_);
            if (err) th->Cancel();
            th->ActiveRelease();
        }

        void LocalActiveRelease()
        {
            ASSERT( m_LocalActiveCount > 0);
            if (sync_dec( &m_LocalActiveCount) == 0)
            {
                    // Send ping to other side to close the connection
                if (m_Remote)
                {
                    SendAsync( RemoteCloseComplete, Wire::MethodCloseIfInactive, "");
                }
                else
                {
                    Cancel();
                    ActiveRelease();
                }
            }
        }

        void /*IFacConfigChannel*/IFacConfigChannel_Release()
        {
            ActiveRelease();
        }

        void CCALL /*remote_receive_i*/release()
        {
            BusyRelease();
        }

        void CCALL /*remote_receive_i*/cancel( int error)
        {
            Cancel();
        }

        void CCALL /*remote_receive_i*/max_sizes(int req_code, size_t* max_input_size, size_t* max_output_size)
        {
        }

        void SendAsync( remote_send_complete_t complete, int code, szc_t format, ...)
        {
            mp_c mp;
            va_list args;
            void* input;
            size_t inputSize;
            va_start( args, format);
            remote_pack_vals_3( &mp, &input, &inputSize, format, &args);
            va_end( args);
            m_Remote->send_async( &m_CancelPin, Channel::m_ConnectTimeout, code, input, inputSize, nullptr, m_MaxOutputSize, complete, this);
        }

        struct ChannelRequestContext
        :
            FacAcx
        {
            remote_req_i* m_Req;
            Channel* m_Channel;
            json_t* m_Input;
            char* m_MethodName;
            json_t* m_Args;

            void* operator new( size_t size, remote_req_i* req)
            {
                C_ASSERT( sizeof( ChannelRequestContext) <= remote_req_scratch_size);
                void* mem = req->scratch();
                memset( mem, 0, sizeof( ChannelRequestContext));
                return mem;
            }

            ChannelRequestContext( remote_req_i* req, Channel* channel)
            :
                FacAcx( &channel->m_CancelPin),
                m_Req( req),
                m_Channel( channel),
                m_Input( nullptr)
            {
                m_Channel->BusyRetain();
            }

            ~ChannelRequestContext()
            {
                m_Channel->BusyRelease();
            }

            void Complete( int err, const void* output, size_t outputSize)
            {
                mp_c mp;
                mp_move( &mp, &m_Mp);
                this->~ChannelRequestContext();
                m_Req->complete( outputSize ? 0 : err, err_to_perr( err), output, outputSize);
            }

            static void Complete( void* rx_, int err, szc_t output, size_t outputSize)
            {
                ChannelRequestContext* rx = SCAST( ChannelRequestContext*, rx_);
                rx->Complete( err, output, outputSize);
            }

            static void ThreadMain( void*);
        };

        void CCALL /*remote_receive_i*/dispatch( remote_req_i* req)
        {
            mp_c mp;
            ChannelRequestContext* rx = new( req) ChannelRequestContext( req, this);
            int err = 0;
            size_t i;
            sz_t serverName;
            sz_t localId;
            int serverRole;
            const void* input = req->input();
            size_t inputSize = req->input_size();
            uint8_t channelPriority[32];
            sz_t peerId, peerPriority;
            int code = rx->m_Req->code();
            switch (code)
            {
            case Wire::MethodIdentify:
                if (pin_i_poll( &m_CancelPin))
                {
                    err = err_broken_pipe;
                    break;
                }
                err = remote_unpack_vals( &mp, input, inputSize, "t32zzzi", channelPriority, &peerId, &serverName, &peerPriority, &serverRole);
                if (err) break;
                if (!sync_xchg( &m_IdentifyPending, true, false))
                {
                        // Cancelled already, or bad peer.
                    break;
                }
                for (i = 0; i < 32; i ++) m_ChannelPriority[i] = m_HalfChannelPriority[i] ^ channelPriority[i];
                m_PeerId = peerId;
                m_ServerName = serverName;
                m_ServerRole = serverRole;
                m_PeerPriority = peerPriority;
                ASSERT( m_IdentifyingCount > 0);
                if (sync_dec( &m_IdentifyingCount) == 0)
                {
                    Connect4();
                    LocalActiveRelease();
                }
                break;
            case Wire::MethodCloseIfInactive:
                if (sync_xchg( &m_RemoteIsActive, true, false))
                {
                    if (m_Peer) m_Peer->ReleaseChannel( this);
                    ActiveRelease();
                }
                break;
            case Wire::Json:
                if (m_IdentifyingCount != 0 || pin_i_poll( &m_CancelPin))
                {
                        // Elect master peer update can hit this easily
                    err = err_access_denied;
                    break;
                }
                ASSERT( m_PeerId);
                rx->m_Channel->m_Config->Dispatch( rx, SCAST( const char*, input));
                rx = nullptr;
                break;
            default:
                err = err_bad_parameter;
                break;
            }
            if (rx) rx->Complete( err, nullptr, 0);
            if (err) Cancel();
        }

        static void PeerConnectedThreadMain( void* ch_)
        {
            Channel* ch = SCAST( Channel*, ch_);
            ch->m_Config->m_Owner->PeerConnected( ch->m_Peer->m_PeerId, ch->m_Peer->m_ServerName, ch->m_Peer->m_ServerRole, ch->m_RemoteAddress, ch->m_LocalAddress);
            ch->BusyRelease();
        }

        void Connect4()
        {
            mp_c mp;
            Peer* peer = nullptr;
            Peer* newPeer = nullptr;
            rblink_t* link = nullptr;
            rblink_t* prev;
            ASSERT( m_PeerId && m_ServerName && m_PeerPriority);
            if (alarm_cancel( &m_ConnectAlarmWire))
            {
                BusyRelease();
            }
            if (m_Config->IsLocalPeer( m_PeerId))
            {
                goto done;
            }
            newPeer = new( t_calloc( Peer)) Peer( m_Config, m_PeerId, m_PeerPriority, m_ServerName);
            m_Config->m_MainLock.lock();
            link = rblink_find_2( m_Config->m_CenterPeer, Peer::Compare, newPeer->m_PeerId.c_str, &prev);
            if (link)
            {
                peer = containerof( Peer, m_PeerLink, link);
                peer->Retain();
            }
            else
            {
                peer = newPeer;
                newPeer = nullptr;
                rblink_insert( &m_Config->m_CenterPeer, prev, &peer->m_PeerLink);
            }
            peer->Retain();
            peer->m_ServerRole = m_ServerRole;
            m_Peer = peer;
            m_Config->m_MainLock.unlock();
            m_Peer->SetActiveChannel( this);
            if (!m_RemoteIsActive) m_Peer->ReleaseChannel( this);
            BusyRetain();
            thread_start_fast( PeerConnectedThreadMain, this);

        done:
            if (newPeer) newPeer->Release();
            if (peer) peer->Release();
        }

        static void CCALL Connect3( void* th_, int inErr, int res, const void* output, size_t outputSize, FD_T unused1[])
        {
            mp_c mp;
            Channel* th = SCAST( Channel*, th_);
            int err = inErr ? inErr : perr_to_err( res);
            if (err) th->Cancel();
            ASSERT( th->m_IdentifyingCount > 0);
            if (sync_dec( &th->m_IdentifyingCount) == 0)
            {
                if (!err) th->Connect4();
                th->LocalActiveRelease();
            }
        }

        static void CCALL Connect2( void* th_, int inErr, ptsst_i* inTransport)
        {
            mp_c mp;
            Channel* th = SCAST( Channel*, th_);
            sz_t localId;
            sz_t priority;
            pt_transport_i* transport = inTransport ? inTransport->as_transport() : nullptr;
            th->BusyRetain();
            remote_receive_i* receiver = th;
            int err = inErr;
//blah
err = err_unsuccessful;
            if (err) goto done;
            th->LocalActiveRetain(); // for identifying count != 0
            th->m_IdentifyingCount = 2; // one for identify pending, one for our identify
            ASSERT( !th->m_IdentifyPending);
            th->m_IdentifyPending = true;
                // nulls out transport and receiver
            VERIFYIS( remote_factory( &th->m_Remote, &th->m_CancelPin, remote_flag_receive | remote_flag_send | remote_flag_keepalive, nullptr, &transport, &receiver), 0);
            th->m_Config->GetLocalIdPriority( &mp, &localId, &priority);
            th->SendAsync( Connect3, Wire::MethodIdentify, "t32zzzi",
                th->m_HalfChannelPriority, localId.c_str, th->m_Config->m_ServerName.c_str, priority.c_str, th->m_Config->m_Owner->GetServerRole());
        done:
            if (err) th->Cancel();
            if (transport) transport->release();
            if (receiver) receiver->release();
            th->LocalActiveRelease();
        }
    };

    void /*IFacConfig*/PeerConnect( FD_T* refFd, bool initiator, szc_t remoteAddress, szc_t localAddress) override
    {
        mp_c mp;
        sz_t localId, groupName;
        ptsec_hasher_i* hasher; ptsec_sha256_hasher_factory( &hasher);
        Channel* channel = nullptr;
        blob_i* secret = nullptr;
        blob_i* hashedSecret = nullptr;
        union { void* v; uint8_t* u8; } dest;
        pt_transport_i* transport = nullptr;
        GetLocalIdPriority( &mp, &localId, nullptr/*priority*/);
        if (localId.isempty()) goto done;
        channel = new( t_calloc( Channel)) Channel( this, initiator, remoteAddress, localAddress);
        ptsec_rng->read( channel->m_HalfChannelPriority, 32);
        m_MainLock.lock();
        secret = m_Secret;
        if (secret) secret->retain();
        groupName = ssu5fold( &mp, m_GroupName);
        m_MainLock.unlock();
        if (!secret) goto done;
        hasher->start();
        hasher->process( groupName, sslen( groupName));
        hasher->process( secret->data(), secret->size());
        hashedSecret = blob_factory_2( 32, &dest.v);
        hasher->finish( dest.u8, 32);
        VERIFYIS( pt_transport_fd_factory( &transport, refFd), 0);
        channel->LocalActiveRetain();
        ptsst_async_factory( &channel->m_CancelPin, Channel::Connect2, channel, &transport, hashedSecret, nullptr, nullptr);
        channel->BusyRetain();
        if (!pin_i_attach( &m_CancelPin, Channel::CancelHandler, &channel->m_CancelWire))
        {
            channel->Cancel();
            channel->BusyRelease();
        }
        channel->BusyRetain();
        if (!alarm_schedule( &channel->m_ConnectAlarmWire, Channel::ConnectAlarmHandler, alarm_now() + 15000))
        {
            channel->Cancel();
            channel->BusyRelease();
        }
        if (pin_i_poll( &channel->m_CancelPin)) channel->Cancel();
        channel->LocalActiveRelease();
        channel = nullptr;
    done:
        if (hasher) hasher->release();
        if (channel)
        {
            channel->Cancel();
            channel->LocalActiveRelease();
        }
        if (transport) transport->release();
        if (secret) secret->release();
        if (hashedSecret) hashedSecret->release();
    }

    Peer* RetainPeer( szc_t peerId)
    {
        Peer* peer = nullptr;
        m_MainLock.lock();
        rblink_t* link = rblink_find( m_CenterPeer, Peer::Compare, peerId);
        if (link)
        {
            peer = containerof( Peer, m_PeerLink, link);
            peer->Retain();
        }
        m_MainLock.unlock();
        return peer;
    }

    size_t/*count*/ RetainPeerListUnderLock( malloc_i* resMa, Peer*** outList)
    {
        ASSERT( m_MainLock.is_locked());
        rblink_t* link = nullptr;
        size_t count = 0;
        size_t peerCount = 0;
        Peer** list;

        while ((link = rblink_next( m_CenterPeer, link)) != nullptr)
        {
            peerCount++;
        }
        list = ma_n_alloc( resMa, Peer*, peerCount + 1);
        link = nullptr;
        while ((link = rblink_next( m_CenterPeer, link)) != nullptr)
        {
            Peer* peer = containerof( Peer, m_PeerLink, link);
            peer->Retain();
            list[count] = peer;
            count++;
        }

        ASSERT( count == peerCount);
        *outList = list;
        return count;
    }

    size_t/*count*/ RetainPeerList( malloc_i* resMa, Peer*** outList)
    {
        Peer** list;
        m_MainLock.lock();
        size_t count = RetainPeerListUnderLock( resMa, &list);
        m_MainLock.unlock();
        *outList = list;
        return count;
    }

        // Scans through peers and picks a master
        // Handles state and initial sync etc...
    Peer* ElectMaster()
    {
        mp_c mp;
        Peer* masterPeer = nullptr;
        Peer* peer;
        Peer** peerList;
        size_t i;
        size_t peerCount = RetainPeerList( &mp, &peerList);
//        if (m_CancelPin.poll()) goto done;
        m_MainLock.lock();
        size_t sepLen = sslen( m_LocalId);
        sz_t masterPriority = sscpy( &mp, m_LocalId.c_str);
        sz_t newPriority = masterPriority;
        sz_t sep;
        for (i = 0; i < peerCount; i++)
        {
            peer = peerList[i];
            if (sscmp( peer->m_PeerPriority, masterPriority) < 0)
            {
                sep = peer->m_PeerPriority;
                while ((sep = ssstr( sep, m_LocalId)) != nullptr)
                {
                    if ((sep.c_str == peer->m_PeerPriority.c_str || sep[-1] == '-') && (sep[sepLen] == 0 || sep[sepLen] == '-'))
                    {
                        break;
                    }
                    sep += sepLen + !!sep[sepLen]; // advance by localid length and separator length
                }
                if (sep) continue;
                masterPeer = peer;
                masterPriority = sscpy( &mp, peer->m_PeerPriority);
            }
        }
        if (masterPeer)
        {
            masterPeer->Retain();
            newPriority = ssformat( &mp, "%s-%s", masterPriority, m_LocalId.c_str);
        }
        Peer* oldMaster = m_MasterPeer;
        if (oldMaster) oldMaster->Retain();
        sz_t oldPriority = sscpy( &mp, m_Priority);
        bool priorityChanged = masterPeer != oldMaster || sscmp( oldPriority, newPriority) != 0;
        if (priorityChanged)
        {
            if (oldMaster && oldMaster->m_Relationship == Peer::Relationship::Master)
            {
                oldMaster->m_Relationship = Peer::Relationship::None;
            }
            m_MasterPeer = masterPeer;
            m_Priority = newPriority;
            if (masterPeer) masterPeer->m_Relationship = Peer::Relationship::Master;
        }
        m_MainLock.unlock();

        if (priorityChanged)
        {
            if (oldMaster)
            {
                oldMaster->SubscribeTo( &mp, nullptr, false);
            }
            if (masterPeer)
            {
                sz_t actualMasterPriority;
                    // SubscribeTo queues ElectMaster on changed priority
                bool subscribed = masterPeer->SubscribeTo( &mp, &actualMasterPriority, true) == 0;
if (!subscribed) tracef( "Subscribe failed 0x%P", masterPeer);
                if (subscribed && !ssisempty( actualMasterPriority) && sscmp( masterPriority, actualMasterPriority) != 0)
                {
tracef( "Subscribe failed due to new master priority 0x%P, %s changed to %s", masterPeer, masterPriority.c_str, actualMasterPriority.c_str);
                    subscribed = false;
                }
                if (!subscribed)
                {
                    m_MainLock.lock();
                    masterPeer->m_Relationship = Peer::Relationship::None;
                    m_MasterPeer = nullptr;
                    m_Priority = newPriority = m_LocalId;
                    m_MainLock.unlock();
                    QueueReplication();
                }
            }
        }
        if (priorityChanged)
        {
                // Tell connected peers to recalculate masters
            for (i = 0; i < peerCount; i++)
            {
                peer = peerList[i];
                peer->UpdatePeerPriority( newPriority);
            }
        }

        if (oldMaster) oldMaster->Release();
        for (i = 0; i < peerCount; i ++) peerList[i]->Release();
        return masterPeer;
    }

    int/*err*/ SyncKey( Peer* peer, sz_t* syncKeyPathList, size_t syncKeyPathCount)
    {
        mp_c mp;
        int err = 0;
        size_t i, j;
        ItemNode** peerItemList;
        ItemNode** ourItemList;
        size_t peerItemCount = peer->ListItems( &mp, &peerItemList, syncKeyPathList, syncKeyPathCount);
        size_t ourItemCount = ListItems( &mp, &ourItemList, syncKeyPathList, syncKeyPathCount);
        FacKey* parentKey = RetainLastKeyFromPath( syncKeyPathList, syncKeyPathCount);

        ASSERT( syncKeyPathCount || (peerItemCount <= 1 && ourItemCount == 1)); // Root key
            // Parent key path should always match the sync key path
//        ASSERT( !syncKeyPath || !parentKey || ss8asciicmpi( parentKey->m_KeyPath, syncKeyPath) == 0);
            // Somehow we ended up on a node that the parent doesn't exist
//        ASSERT( !!parentKey == !syncKeyPath.isempty());
            // Root has no vals
//        ASSERT( !parentKey || ss8asciicmpi( parentKey->m_KeyPath, syncKeyPath) == 0);
            // No need to retain/release these as they are handled under lock only
            // Check for newer vals and null them out after syncing
        for (i = 0; i < peerItemCount; i++)
        {
            bool update = true;
            // bool deleteItem = false;
            ItemNode* peerItem = peerItemList[i];
            for (j = 0; j < ourItemCount; j++)
            {
                ItemNode* ourItem = ourItemList[j];
                if (!ourItem) continue;

                if (ss8asciicmpi( peerItem->name, ourItem->name) != 0) continue;

                ASSERT( parentKey || sscmp( ourItem->name, "root") == 0);
                if (ourItem->timestamp < peerItem->timestamp)
                {
                    ourItemList[j] = nullptr;
                    if (ourItem->type != peerItem->type)
                    {
                        // deleteItem = true;
                    }
                }
                else if (ourItem->timestamp == peerItem->timestamp)
                {
                    if (ourItem->type == peerItem->type)
                    {
                        ourItemList[j] = nullptr;
                        update = false;
                    }
                    else
                    {
                            // Take peer's item and delete ours
                        ourItemList[j] = nullptr;
                        // deleteItem = true;
                    }
                }
                else if (ourItem->timestamp > peerItem->timestamp)
                {
                    ourItem = nullptr;
                    update = false;
                }
                break;
            }

                // Change our item
            if (update)
            {
                ASSERT( peerItem->type == FacItem::valType || peerItem->type == FacItem::folderType || peerItem->type == FacItem::whiteoutType);
                if (peerItem->type == FacItem::whiteoutType)
                {
                    int64_t timestamp = peerItem->timestamp;
                    parentKey->DeleteItem( peerItem->name, !FacWhiteout::Expired( timestamp)/*leaveWhiteout*/, false/*whiteoutOnMatchOnly*/, timestamp, false/*propagateTimestamp*/, peer->m_PeerId);
                }
                if (peerItem->type == FacItem::valType)
                {
                    parentKey->SetVal( peerItem->name, peerItem->val, peerItem->timestamp, false/*propagateTimestamp*/, peer->m_PeerId);
                }
                else if (peerItem->type == FacItem::folderType)
                {
                    FacKey* key = nullptr;
                    if (parentKey)
                    {
                            // Create temp key with old timestamp to trigger update if sync is interrupted
                        key = parentKey->CreateKey( peerItem->name, 0, false/*propagateTimestamp*/, false/*overwriteTimestamp*/, peer->m_PeerId);
                        if (key) key->Release();
                    }
                    sz_t* keyPathList;
                    size_t keyPathCount = InitKeyPathList( &mp, &keyPathList, syncKeyPathList, syncKeyPathCount, peerItem->name);
                    int check = SyncKey( peer, keyPathList, keyPathCount);
                    if (parentKey)
                    {
                        if (!check)
                        {
                                // Set the key's timestamp the peer's timestamp
                            key = parentKey->CreateKey( peerItem->name, peerItem->timestamp, false/*propagateTimestamp*/, true/*overwriteTimestamp*/, peer->m_PeerId);
                            if (key) key->Release();
                        }
                    }
                }
            }
        }

            // Any non-null vals don't exist on the peer
        for (i = 0; i < ourItemCount; i++)
        {
            ItemNode* ourItem = ourItemList[i];
            if (!ourItem) continue;
            ASSERT( parentKey || sscmp( ourItem->name, "root") == 0);
            ASSERT( ourItem->type == FacItem::valType || ourItem->type == FacItem::folderType || ourItem->type == FacItem::whiteoutType);
            if (ourItem->type == FacItem::valType)
            {
                peer->SetVal( syncKeyPathList, syncKeyPathCount, ourItem->name, ourItem->val, ourItem->timestamp, false/*propagateTimestamp*/);
            }
            else if (ourItem->type == FacItem::whiteoutType)
            {
                peer->DeleteItem( syncKeyPathList, syncKeyPathCount, ourItem->name, !FacWhiteout::Expired( ourItem->timestamp)/*leaveWhiteout*/, ourItem->timestamp, false/*propagateTimestamp*/);
            }
            else if (ourItem->type == FacItem::folderType)
            {
                sz_t* keyPathList;
                size_t keyPathCount = InitKeyPathList( &mp, &keyPathList, syncKeyPathList, syncKeyPathCount, ourItem->name);
                    // Create temp key with old timestamp to trigger update if sync is interrupted
                int check = peer->CreateKey( keyPathList, keyPathCount, 0, false/*propagateTimestamp*/, false/*overwriteTimestamp*/);
                if (!check)
                {
                    check = SyncKey( peer, keyPathList, keyPathCount);
                    if (!check)
                    {
                            // Set the key's timestamp the our timestamp
                        check = peer->CreateKey( keyPathList, keyPathCount, ourItem->timestamp, false/*propagateTimestamp*/, true/*overwriteTimestamp*/);
                    }
                }
            }
            else
            {
                ASSERT( "Unknown item type" && false);
            }
        }

        if (parentKey) parentKey->Release();
        return err;
    }

    void SyncWithPeer( Peer* masterPeer)
    {
            // Sync root
        SyncKey( masterPeer, nullptr/*keyPathList*/, 0/*keyPathCount*/);
    }

    static void ReplicationNeededThreadMain( void* config_)
    {
        FacConfig* config = SCAST( FacConfig*, config_);
        mp_c mp;
        Peer* masterPeer;
        for (;;)
        {
            sync_xchg( &config->m_ReplicationNeeded, true, false);

            masterPeer = config->ElectMaster();
            if (masterPeer)
            {
                config->SyncWithPeer( masterPeer);
                masterPeer->Release();
            }

            VERIFY( sync_xchg( &config->m_ReplicationInProgress, true, false));
            if (!config->m_ReplicationNeeded) break;
            if (!sync_xchg( &config->m_ReplicationInProgress, false, true)) break;
        }
        config->BusyRelease();
    }

    void QueueReplication()
    {
        m_ReplicationNeeded = true;
        if (sync_xchg( &m_ReplicationInProgress, false, true))
        {
            BusyRetain();
            thread_start_fast( ReplicationNeededThreadMain, this);
        }
    }

    size_t/*count*/ /*IFacConfig*/GetServerInfoList( malloc_i* resMa, FacAPI::ServerInfo*** outList) override
    {
        mp_c mp;
        json_t* resultsObj;
        json_t* args = ptjson_setv( &mp, ptjson_new_object( &mp),
            pjv( "method", "get_server_info"),
            pjv( "args", ptjson_new_object( &mp)),
            pjv_end());
        RelayHub( &mp, &resultsObj, nullptr/*peerId*/, args, 0/*hopCount*/);
        json_t** values;
        size_t count = ptjson_get_list( &mp, nullptr, &values, resultsObj);
        FacAPI::ServerInfo** list = ma_n_alloc( resMa, FacAPI::ServerInfo*, count + 1);
        size_t i = 0;
        json_t* val;
        json_t* jsonInfo;
        for (i = 0; i < count; i++)
        {
            val = values[i];
            ptjson_getv( &mp, val, pjv( "info", &jsonInfo), pjv_end());
            list[i] = UnpackServerInfo( resMa, jsonInfo);
        }
        ASSERT( count == i);
        *outList = list;
        return count;
    }

    size_t /*IFacConfig*/GetServerNameList( malloc_i* resMa, sz_t** outList) override
    {
        mp_c mp;
        Peer** peerList;
        size_t peerCount = RetainPeerList( &mp, &peerList);
        sz_t* list = ma_n_alloc( resMa, sz_t, peerCount + 2);
        size_t i;
        for (i = 0; i < peerCount; i++)
        {
            Peer* peer = peerList[i];
            list[i+1] = sscpy( resMa, peer->m_ServerName);
            peer->Release();
        }
        list[0] = sscpy( resMa, m_ServerName);
        peerCount++;
        *outList = list;
        return peerCount;
    }

    void /*IFacConfig*/GetServerGroupMaster( malloc_i* resMa, sz_t* outServerMasterUid) override
    {
        sz_t serverMasterUid = nullptr;

        m_MainLock.lock();
        Peer* masterPeer = m_MasterPeer;
        if (masterPeer)
        {
            serverMasterUid = sscpy( resMa, masterPeer->m_PeerId);
        }
        m_MainLock.unlock();

        *outServerMasterUid = serverMasterUid;
    }

    void /*IFacConfig*/GetServerGroupName( malloc_i* resMa, sz_t* outServerGroupName) override
    {
        m_MainLock.lock();
        sz_t serverGroupName = sscpy( resMa, m_GroupName);
        m_MainLock.unlock();
        *outServerGroupName = serverGroupName;
    }

    void SendPeerProxyRequest( pin_i* cancelPin, FacRequestComplete complete, void* context, szc_t serverUid, szc_t methodName, szc_t input)
    {
        Peer* peer = RetainPeer( serverUid);
        if (!peer)
        {
            if (complete) complete( context, err_broken_pipe, nullptr, 0);
        }
        else
        {
            peer->SendProxyRequest( cancelPin, complete, context, methodName, input);
            peer->Release();
        }
    }

    void /*IFacConfig*/PeerHubRequest( pin_i* cancelPin, FacRequestComplete complete, void* context, szc_t serverUid, szc_t input) override
    {
        SendPeerProxyRequest( cancelPin, complete, context, serverUid, "hub_request", input);
    }

    void /*IFacConfig*/PeerStorageRequest( pin_i* cancelPin, FacRequestComplete complete, void* context, szc_t serverUid, szc_t input) override
    {
        SendPeerProxyRequest( cancelPin, complete, context, serverUid, "storage_request", input);
    }

    void /*IFacConfig*/PeerRaidEngineRequest( pin_i* cancelPin, FacRequestComplete complete, void* context, szc_t serverUid, szc_t input) override
    {
        SendPeerProxyRequest( cancelPin, complete, context, serverUid, "dre_request", input);
    }

        // Pushes changes to peers
    void RepNotify()
    {
        mp_c mp;
        Peer** peerList;
        size_t i;
        size_t peerCount = RetainPeerList( &mp, &peerList);
        RepChange* repChange;
        if (!peerCount)
        {
            tracef( "Skipping replication because no peers");
        }
        while ((repChange = PopRepChange()) != nullptr)
        {
//if (peerCount) tracef( "Running through replication \"%s/%s\" - %s - %i", repChange->keyPath.c_str, repChange->name.c_str, repChange->owningPeerId.c_str, repChange->type);

            for (i = 0; i < peerCount; i++)
            {
                Peer* peer = peerList[i];
                int relationship = peer->GetRelationship();
                if (ss8asciicmpi( peer->m_PeerId, repChange->owningPeerId) != 0 &&
                    relationship != Peer::Relationship::None)
                {
//tracef( "Replicating key \"%s\" to peer \"%s\"", repChange->keyPath.c_str, peer->m_PeerId.c_str);
                    if (repChange->type == FacItem::folderType)
                    {
                        int check = peer->CreateKey( repChange->keyPathList, repChange->keyPathCount, repChange->timestamp, true/*propagateTimestamp*/, true/*overwriteTimestamp*/);
                        if (check)
                        {
//tracef( "Failed to replicate key");
                        }
                    }
                    else if (repChange->type == FacItem::valType)
                    {
//tracef( "Replicating val \"%s/%s\" to peer \"%s\"", repChange->keyPath.c_str, repChange->name.c_str, peer->m_PeerId.c_str);
                        int check = peer->SetVal( repChange->keyPathList, repChange->keyPathCount, repChange->name, repChange->val, repChange->timestamp, false/*propagateTimestamp*/);
                        if (check)
                        {
//tracef( "Failed to replicate val");
                        }
                    }
                    else if (repChange->type == FacItem::whiteoutType)
                    {
//tracef( "Replicating whiteout \"%s/%s\" to peer \"%s\"", repChange->keyPath.c_str, repChange->name.c_str, peer->m_PeerId.c_str);
                        int check = peer->DeleteItem( repChange->keyPathList, repChange->keyPathCount, repChange->name, true/*leaveWhiteout*/, repChange->timestamp, false/*propagateTimestamp*/);
                        if (check)
                        {
//tracef( "Failed to replicate val");
                        }
                    }
                }
            }

            repChange->Release();
            if (m_CancelPin.poll()) break;
        }
        for (i = 0; i < peerCount; i++)
        {
            Peer* peer = peerList[i];
            peer->Release();
        }
    }

    void /*IFacConfig*/Configure( szc_t identifier, szc_t groupName, blob_i* inSharedSecret) override
    {
        mp_c mp;
        blob_i* sharedSecret = groupName.isempty() ? nullptr : inSharedSecret;
        if (sharedSecret) sharedSecret->retain();
        blob_i* extra = nullptr;
        Peer** peerList;
        size_t i;

        m_MainLock.lock();
        m_GroupName = groupName;
        m_Priority = identifier;
        m_LocalId = identifier;
        if (m_Secret) extra = m_Secret;
        m_Secret = sharedSecret;
        size_t peerCount = RetainPeerListUnderLock( &mp, &peerList);
        m_MainLock.unlock();

        for (i = 0; i < peerCount; i++)
        {
            Peer* peer = peerList[i];
            peer->Cancel();
            peer->Release();
        }

        QueueReplication();
        if (extra) extra->release();
    }

    void IFacConfigAPI_Release()
    {
        Release();
    }

    size_t/*count*/ RetainKeysFromPath( malloc_i* resMa, FacKey*** outList, sz_t* keyPathList, size_t keyPathCount)
    {
        mp_c mp;
        size_t count = 0;
        FacKey** list = ma_n_alloc( resMa, FacKey*, keyPathCount + 1);
        FacKey* prevKey = nullptr;
        size_t i;
        for (i = 0; i < keyPathCount; i++)
        {
            FacKey* key;
            if (i == 0)
            {
                ASSERT( sscmp( keyPathList[i], root->m_Name) == 0);
                key = root;
                root->Retain();
            }
            else
            {
                key = prevKey->RetainKey( keyPathList[i]);
            }
            if (key)
            {
                list[i] = key;
                ASSERT( ss8asciicmpi( keyPathList[i], key->m_Name) == 0);
                count = count + 1;
                prevKey = key;
            }
            else
            {
                break;
            }
        }
        ASSERT( count <= keyPathCount);
        *outList = list;
        return count;
    }

    size_t/*count*/ CreateKeysFromPath( malloc_i* resMa, FacKey*** outList, sz_t* keyPathList, size_t keyPathCount,
        int64_t timestamp, bool propagateTimestamp, bool overwriteTimestamp, szc_t owningPeerId)
    {
        mp_c mp;
        size_t count = 0;
        FacKey** list = ma_n_alloc( resMa, FacKey*, keyPathCount + 1);
        FacKey* prevKey = nullptr;
        size_t i;
        for (i = 0; i < keyPathCount; i++)
        {
            FacKey* key;
            if (i == 0)
            {
                ASSERT( sscmp( keyPathList[i], root->m_Name) == 0);
                key = root;
                root->Retain();
            }
            else if (i == keyPathCount - 1)
            {
                key = prevKey->CreateKey( keyPathList[i], timestamp, propagateTimestamp, overwriteTimestamp, owningPeerId);
            }
            else
            {
                key = prevKey->CreateKey( keyPathList[i], 0, false, false, owningPeerId);
            }
            if (key)
            {
                list[i] = key;
                ASSERT( ss8asciicmpi( keyPathList[i], key->m_Name) == 0);
                count = count + 1;
                prevKey = key;
            }
            else
            {
                break;
            }
        }
        ASSERT( count <= keyPathCount);
        *outList = list;
        return count;
    }

    FacKey* CreateLastKeyFromPath( sz_t* keyPathList, size_t keyPathCount, int64_t timestamp, bool propagateTimestamp, bool overwriteTimestamp, szc_t owningPeerId)
    {
        mp_c mp;
        sz_t lastKeyName;
        FacKey* key = nullptr;
        size_t i;
        FacKey** list;
        size_t count = CreateKeysFromPath( &mp, &list, keyPathList, keyPathCount,
            timestamp, propagateTimestamp, overwriteTimestamp, owningPeerId);
        for (i = 0; i < count; i++)
        {
            if (i == keyPathCount - 1)
            {
                key = list[i];
                key->Retain();
            }
            list[i]->Release();
        }
        return key;
    }

    FacKey* RetainLastKeyFromPath( sz_t* keyPathList, size_t keyPathCount)
    {
        mp_c mp;
        sz_t lastKeyName;
        FacKey* key = nullptr;
        size_t i;
        FacKey** list;
        size_t count = RetainKeysFromPath( &mp, &list, keyPathList, keyPathCount);
        for (i = 0; i < count; i++)
        {
            if (i == keyPathCount - 1)
            {
                key = list[i];
                key->Retain();
            }
            list[i]->Release();
        }
        return key;
    }

    int/*err*/ SetValFromKeyPath( sz_t* keyPathList, size_t keyPathCount, szc_t name, szc_t val, int64_t timestamp, bool propagateTimestamp, szc_t owningPeerId)
    {
        FacKey* parentKey = CreateLastKeyFromPath( keyPathList, keyPathCount, timestamp, propagateTimestamp, false/*overwriteTimestamp*/, owningPeerId);
        int err = parentKey ? 0 : err_invalid_parameter;
        if (!err)
        {
            err = parentKey->SetVal( name, val, timestamp, propagateTimestamp, owningPeerId);
            parentKey->Release();
        }
        return err;
    }

    int/*err*/ CreateKeyFromKeyPath( sz_t* keyPathList, size_t keyPathCount, int64_t timestamp, bool propagateTimestamp, bool overwriteTimestamp, szc_t owningPeerId)
    {
        mp_c mp;
        int err = err_invalid_parameter;
        if (ss8asciicmpi( root->m_Name, keyPathList[ keyPathCount - 1]) == 0)
        {
            if (overwriteTimestamp)
            {
                root->UpdateTimestamp( timestamp, propagateTimestamp);
            }
            err = 0;
        }
        else
        {
            FacKey* parentKey = CreateLastKeyFromPath( keyPathList, keyPathCount, timestamp, propagateTimestamp, overwriteTimestamp, owningPeerId);
            if (parentKey)
            {
                err = 0;
                parentKey->Release();
            }
        }
        return err;
    }

    int/*err*/ DeleteItemFromKeyPath( sz_t* keyPathList, size_t keyPathCount, szc_t name, bool leaveWhiteout, int64_t timestamp, bool propagateTimestamp, szc_t owningPeerId)
    {
        mp_c mp;
        int err = err_invalid_parameter;
        ASSERT( keyPathCount);
        ASSERT( ss8asciicmpi( name, "root") != 0);
        FacKey* parentKey = CreateLastKeyFromPath( keyPathList, keyPathCount, timestamp, propagateTimestamp, false/*overwriteTimestamp*/, owningPeerId);
        if (parentKey)
        {
            err = parentKey->DeleteItem( name, leaveWhiteout, false/*whiteoutOnMatchOnly*/, timestamp, propagateTimestamp, owningPeerId);
            parentKey->Release();
        }
        return err;
    }

    size_t/*count*/ ListItems( malloc_i* resMa, ItemNode*** outList, sz_t* keyPathList, size_t keyPathCount)
    {
        ItemNode** nodeList = ma_n_alloc( resMa, ItemNode*, 1);
        size_t count = 0;
        size_t i;
        if (!keyPathCount)
        {
            nodeList = ma_n_alloc( resMa, ItemNode*, 2);
            nodeList[0] = ItemNodeFactory( resMa, root);
            count = 1;
        }
        else
        {
            FacKey* key = RetainLastKeyFromPath( keyPathList, keyPathCount);
            if (key)
            {
                key->m_KeyLock.lock();
                rblink_t* link = nullptr;
                while ((link = rblink_next( key->m_Center, link)) != nullptr)
                {
                    count++;
                }
                nodeList = ma_n_alloc( resMa, ItemNode*, count + 1);
                link = nullptr;
                i = 0;
                while ((link = rblink_next( key->m_Center, link)) != nullptr)
                {
                    FacItem* item = containerof( FacItem, m_Link, link);
                    nodeList[i] = ItemNodeFactory( resMa, item);
                    i++;
                }
                key->m_KeyLock.unlock();
                key->Release();
                ASSERT( i == count);
            }
        }
        (*outList) = nodeList;
        return count;
    }

    int/*err*/ PeerCanRepNotify( szc_t peerId)
    {
        Peer* peer = RetainPeer( peerId);
        int err = peer ? 0 : err_not_found;
        if (!err)
        {
            int relationship = peer->GetRelationship();
            if (relationship == Peer::Relationship::None)
            {
                err = err_access_denied;
tracef( "Peer: %s trying to communicate when not a slave or master", peerId.c_str);
            }
            peer->Release();
        }
        else
        {
tracef( "Peer: No connection to a peer with id %s", peerId.c_str);
        }
        return err;
    }

    int/*err*/ SetVal( malloc_i* ma, szc_t peerId, json_t* request, json_t** outResponse)
    {
        mp_c mp;
        sz_t* keyPathList;
        szc_t name;
        szc_t val;
        int64_t timestamp;
        int propagateTimestamp;
        json_t* keyPathJsonList;
        ptjson_getv( &mp, request,
            pjv( "key_path_list"      , &keyPathJsonList   ),
            pjv( "name"               , &name              ),
            pjv( "val"                , &val               ),
            pjv( "timestamp"          , &timestamp         ),
            pjv( "propagate_timestamp", &propagateTimestamp),
            pjv_end());
        size_t keyPathCount = UnpackStringList( &mp, &keyPathList, keyPathJsonList);
        int err = SetValFromKeyPath( keyPathList, keyPathCount, name, val, timestamp, !!propagateTimestamp, peerId);
        return err;
    }

    int/*err*/ CreateKey( malloc_i* ma, szc_t peerId, json_t* request, json_t** outResponse)
    {
        mp_c mp;
        sz_t* keyPathList;
        int64_t timestamp;
        int propagateTimestamp;
        int overwriteTimestamp;
        json_t* keyPathJsonList;
        ptjson_getv( &mp, request,
            pjv( "key_path_list"      , &keyPathJsonList   ),
            pjv( "timestamp"          , &timestamp         ),
            pjv( "propagate_timestamp", &propagateTimestamp),
            pjv( "overwrite_timestamp", &overwriteTimestamp),
            pjv_end());
        size_t keyPathCount = UnpackStringList( &mp, &keyPathList, keyPathJsonList);
        int err = CreateKeyFromKeyPath( keyPathList, keyPathCount, timestamp, !!propagateTimestamp, !!overwriteTimestamp, peerId);
        return err;
    }

    int/*err*/ DeleteItem( malloc_i* ma, szc_t peerId, json_t* request, json_t** outResponse)
    {
        mp_c mp;
        sz_t* keyPathList;
        int64_t timestamp;
        szc_t name;
        int leaveWhiteout;
        int propagateTimestamp;
        json_t* keyPathJsonList;
        ptjson_getv( &mp, request,
            pjv( "key_path_list"      , &keyPathJsonList   ),
            pjv( "name"               , &name              ),
            pjv( "leave_whiteout"     , &leaveWhiteout     ),
            pjv( "timestamp"          , &timestamp         ),
            pjv( "propagate_timestamp", &propagateTimestamp),
            pjv_end());
        size_t keyPathCount = UnpackStringList( &mp, &keyPathList, keyPathJsonList);
        int err = DeleteItemFromKeyPath( keyPathList, keyPathCount, name, !!leaveWhiteout, timestamp, !!propagateTimestamp, peerId);
        return err;
    }

    int/*err*/ SubscribeToMaster( malloc_i* ma, szc_t peerId, json_t* request, json_t** outResponse)
    {
        mp_c mp;
        int toggle;
        ptjson_getv( &mp, request,
            pjv( "toggle", &toggle),
            pjv_end());
        Peer* peer = RetainPeer( peerId);
        int err = peer ? 0 : err_not_found;
        m_MainLock.lock();
        szc_t masterPriority = sscpy( &mp, m_Priority);
        if (peer && peer != m_MasterPeer)
        {
            peer->m_Relationship = toggle ? Peer::Relationship::Slave : Peer::Relationship::None;
        }
        m_MainLock.unlock();
        if (peer) peer->Release();
        *outResponse = ptjson_setv( ma, ptjson_new_object( ma),
            pjv( "master_priority", masterPriority),
            pjv_end());
        return err;
    }

    int/*err*/ ListItems( malloc_i* ma, szc_t peerId, json_t* request, json_t** outResponse)
    {
        mp_c mp;
        json_t* keyPathJsonList;
        ptjson_getv( &mp, request,
            pjv( "key_path_list", &keyPathJsonList),
            pjv_end());
        sz_t* keyPathList;
        size_t keyPathCount = UnpackStringList( &mp, &keyPathList, keyPathJsonList);
        ItemNode** list;
        size_t count = ListItems( &mp, &list, keyPathList, keyPathCount);
        json_t* jsonList = ptjson_new_array( ma);
        size_t i;
        for (i = 0; i < count; i++)
        {
            ItemNode* node = list[i];
            ptjson_set_array( ma, jsonList, PackItemNode( ma, node));
        }
        *outResponse = ptjson_setv( ma, ptjson_new_object( ma),
            pjv( "list", jsonList),
            pjv_end());
        return 0;
    }

    int/*err*/ UpdatePeerPriority( malloc_i* ma, szc_t peerId, json_t* request, json_t** outResponse)
    {
        mp_c mp;
        int err = 0;
        szc_t priority;
        ptjson_getv( &mp, request,
            pjv( "priority", &priority.c_str),
            pjv_end());
        Peer* peer = RetainPeer( peerId);
            // This can happen when rejoining server groups, peers are torn down forcefully
        // ASSERT( peer); // Race?
        if (peer)
        {
                // SetPriority searches for new master
            peer->SetPriority( priority);
            peer->Release();
        }
        else
        {
            err = err_not_found;
        }
        return err;
    }

    int/*err*/ RelayHub( malloc_i* ma, szc_t peerId, json_t* request, json_t** outResponse)
    {
        mp_c mp;
        szc_t methodName;
        json_t* args;
        size_t hopCount;
        ptjson_getv( &mp, request,
            pjv( "args"     , &args    ),
            pjv( "hop_count", &hopCount),
            pjv_end());
        json_t* resultsObj;
        int err = RelayHub( ma, &resultsObj, peerId, args, hopCount);
        *outResponse = ptjson_setv( ma, ptjson_new_object( ma),
            pjv( "results", resultsObj),
            pjv_end());
        return err;
    }

#define FACCONFIG_API_METHOD_LIST \
    X_( set_val             , SetVal            ) \
    X_( create_key          , CreateKey         ) \
    X_( delete_item         , DeleteItem        ) \
    X_( subscribe_to_master , SubscribeToMaster ) \
    X_( list_items          , ListItems         ) \
    X_( update_peer_priority, UpdatePeerPriority) \
    X_( relay_hub           , RelayHub          ) \

    void Dispatch( Channel::ChannelRequestContext*, szc_t input);

    typedef int/*err*/ (*JsonHandler)(FacConfig* config, malloc_i* ma, szc_t peerId, json_t* args, json_t** outResponse);
    #define X_(A,B) static int/*err*/ B##_(FacConfig* th, malloc_i* ma, szc_t peerId, json_t* args, json_t** outResponse) { return th->B( ma, peerId, args, outResponse); }
    FACCONFIG_API_METHOD_LIST
    #undef X_
    struct Method
    {
        szc_t Name;
        JsonHandler Handler;
    };
    static Method Methods[];
};

FacConfig::Method FacConfig::Methods[] =
{
#define X_(A,B) { #A, &B##_ },
FACCONFIG_API_METHOD_LIST
#undef X_
};

void FacConfig::Channel::ChannelRequestContext::ThreadMain( void* rx_)
{
    mp_c mp;
    ChannelRequestContext* rx = SCAST( ChannelRequestContext*, rx_);
    int err = 0;
    json_t* output = nullptr;
    size_t i = 0;
    size_t outputSize;
    sz_t outputText;
    for (;;)
    {
        if (i == countof( FacConfig::Methods))
        {
            err = err_bad_parameter;
            break;
        }
        if (ssasciicmpi( rx->m_MethodName, FacConfig::Methods[i].Name) == 0)
        {
            err = FacConfig::Methods[i].Handler( rx->m_Channel->m_Config, &mp, rx->m_Channel->m_PeerId, ptjson_new_copy( &mp, rx->m_Args), &output);
            break;
        }
        i ++;
    }
    if (err)
    {
        if (!output) output = ptjson_new_object( &mp);
        ptjson_set_int( &mp, output, "err", err_to_perr( err));
    }
    outputText = ptjson_export( &mp, &outputSize, output, false);
    rx->Complete( err, outputText, outputSize);
}

void FacConfig::Dispatch( Channel::ChannelRequestContext* rx, szc_t input)
{
    mp_c mp;
    FacConfig* config = rx->m_Channel->m_Config;
    rx->m_Input =  ptjson_import( &rx->m_Mp, nullptr, input, SIZE_MAX);
    ptjson_getv( &rx->m_Mp, rx->m_Input,
        pjv( "method", &rx->m_MethodName),
        pjv( "args"  , &rx->m_Args      ),
        pjv_end());

    bool hubRequest = ssasciicmpi( rx->m_MethodName, "hub_request") == 0;
    bool storageRequest = !hubRequest && ssasciicmpi( rx->m_MethodName, "storage_request") == 0;
    if (hubRequest || storageRequest || ssasciicmpi( rx->m_MethodName, "dre_request") == 0)
    {
        json_t* subInput;
        ptjson_getv( &mp, rx->m_Args,
            pjv( "request", &subInput),
            pjv_end());
        sz_t subInputText = ptjson_export( &mp, nullptr, subInput, false);
        if (hubRequest)
        {
            config->m_Owner->HubRequest( rx, Channel::ChannelRequestContext::Complete, rx, subInputText);
        }
        else if (storageRequest)
        {
            config->m_Owner->StorageRequest( rx, Channel::ChannelRequestContext::Complete, rx, subInputText);
        }
        else
        {
            config->m_Owner->RaidEngineRequest( rx, Channel::ChannelRequestContext::Complete, rx, subInputText);
        }
    }
    else
    {
        thread_start_fast( Channel::ChannelRequestContext::ThreadMain, rx);
    }
}

void CopyFacKey( IFacKey* source, IFacKey* dest)
{
    mp_c mp;
    char* jsonStr = source->ConvertToJson( &mp);
    json_t* json = ptjson_import( &mp, nullptr, jsonStr, sslen( jsonStr));
    dest->LoadJson( json);
}

struct CommitThreadParam
{
    FacConfig* m_Config;

    CommitThreadParam( FacConfig* c)
    :
        m_Config( c)
    {
        m_Config->BusyRetain();
    }

    ~CommitThreadParam()
    {
        m_Config->BusyRelease();
    }
};

static void CommitThreadMain( void* config_)
{
    FacConfig* config = SCAST( FacConfig*, config_);
    {
        pin_i* cancelPin = config->CancelPin();
        pin_i* commitPin = config->CommitPin();
        pin_i* refreshPin = procsig_pin(FILE_SIGNAL_REFRESH);
            // Delay processing of commit pin to avoid thrashing.
        static const int commitDelay = 5000;
        pin_c delayedCommitPin;
        pin_delayed_relay_c commitPinRelay( commitPin, commitDelay, &delayedCommitPin);

        while (!cancelPin->poll())
        {
            pin_buf_x_t pinBuf;
            pin_buf_2_init( &pinBuf, refreshPin, &delayedCommitPin);
            pin_wait( pin_buf_x_output( &pinBuf), cancelPin, -1);
            if (refreshPin->poll())
            {
                // Any refresh stuff here
            }
            if (delayedCommitPin.poll())
            {
                delayedCommitPin.reset();
                commitPinRelay.reset( commitDelay);
                config->Commit();
            }
            pin_buf_x_term(&pinBuf);
        }
    }
    config->BusyRelease();
}

static void RepNotifyThreadMain( void* config_)
{
    FacConfig* config = SCAST( FacConfig*, config_);
    {
        pin_i* cancelPin = config->CancelPin();
        pin_i* replicationPin = config->ReplicationPin();
            // Delay processing of replication pin to avoid thrashing.
        static const int replicationDelay = 1000;
        pin_c delayedReplicationPin;
        pin_delayed_relay_c replicationPinRelay( replicationPin, replicationDelay, &delayedReplicationPin);

        while (!cancelPin->poll())
        {
            pin_wait( &delayedReplicationPin, cancelPin, -1);
            if (delayedReplicationPin.poll())
            {
                config->RepNotify();
                delayedReplicationPin.reset();
                replicationPinRelay.reset( replicationDelay);
            }
        }
    }
    config->BusyRelease();
}

IFacKey* FacKey_OpenPath( IFacKey* rootKey, const szc_t* keyNameList, size_t keyNameCount)
{
    size_t i = 0;
    szc_t name;
    IFacKey* parent = nullptr;
    IFacKey* key = rootKey;
    if (!key) goto done;
    key->Retain();
    while (i < keyNameCount)
    {
        if (parent) parent->Release();
        parent = key;
        key = nullptr;
        name = keyNameList[ i ++];
        if (ssisempty( name)) goto done;
        key = parent->OpenKey( name);
        if (!key) goto done;
    }
done:
    if (parent) parent->Release();
    return key;
}

IFacKey* FacKey_CreatePath( IFacKey* rootKey, const szc_t* keyNameList, size_t keyNameCount)
{
    size_t i = 0;
    szc_t name;
    IFacKey* parent = nullptr;
    IFacKey* key = rootKey;
    if (!key) goto done;
    key->Retain();
    while (i < keyNameCount)
    {
        if (parent) parent->Release();
        parent = key;
        key = nullptr;
        name = keyNameList[ i ++];
        if (ssisempty( name)) goto done;
        key = parent->CreateKey( name);
        if (!key) goto done;
    }
done:
    if (parent) parent->Release();
    return key;
}

IFacConfig* FacConfigFactory( IFacOutput* output, IFacConfigOwner* owner, IHubAPI* hub, szc_t storageLocation, szc_t serverName, int flags)
{
    TRACE_OPEN( "facconfig");
    FacConfig* config = new( t_calloc( FacConfig)) FacConfig( output, owner, hub, storageLocation);
    config->m_ServerName = serverName;
    config->root = config->KeyFactory( "root");
    config->root->InitKeyPath();
    config->LoadConfig();
    if (config->m_FailedLoad)
    {
        config->SaveBackup();
    }
    if (!(flags & FacConfigFlags::IgnoreAutoSave))
    {
        config->BusyRetain();
        thread_start_fast( CommitThreadMain, config);
    }
    config->BusyRetain();
    thread_start_fast( RepNotifyThreadMain, config);
    config->m_Started = true;
    return config;
}
